define('feature/comments/activity-comment-context', [
    'jquery',
    'underscore',
    'eve',
    'util/client-storage',
    'widget/markup-preview',
    'widget/mentionable-textarea',
    'feature/comments/comment-container'
],
function (
    $,
    _,
    eve,
    clientStorage,
    markupPreview,
    MentionableTextarea,
    CommentContainer
) {

    return CommentContainer.extend({
        initialize: function() {
            CommentContainer.prototype.initialize.apply(this, arguments);

            var draft = clientStorage.getSessionItem(this.getDraftKey());
            if (draft && draft.text) {
                //Backbone events (used to trigger the reply/edit forms) don't get bound until AFTER initialise is called, so wait for that to happen.
                _.defer(_.bind(this.restoreDraft, this, draft));
            }

            this.mentionableTextarea = new MentionableTextarea({
                selector: '.general-comment-activity textarea, .general-comment-form textarea',
                $container: this.$el
            });
        },
        rootCommentListSelector : '.pull-request-activity',
        events : { // override the events so we're not handling diff comment events. Only non-diff comments.
            'submit .general-comment-activity form, form.general-comment-form' : 'onFormSubmit',
            'click .general-comment-activity .cancel, form.general-comment-form .cancel' : 'onCancelClicked',
            'click .general-comment-activity .reply' : 'onReplyClicked',
            'click .general-comment-activity .edit' : 'onEditClicked',
            'keydown .general-comment-activity textarea, .general-comment-form textarea' : 'onTextareaKeydown',
            'focus .general-comment-form textarea' : 'onGeneralFormTextareaFocused'
        },
        initDeleteButtons : function() {
            this.createDeleteDialog().attachTo('.general-comment-activity .delete', null, this.el);
        },
        insertCommentIntoList : function($comment, $commentList, commentJSON) {
            if ($commentList.is(this.rootCommentListSelector)) {
                // TODO: we need to order it along with other activity items.
                // Luckily, until we do activity reloading, we can be assured we're only adding comments at the top.

                var $generalCommentForm = $commentList.children(':first');
                $comment.insertAfter($generalCommentForm);
            } else {
                CommentContainer.prototype.insertCommentIntoList.apply(this, arguments);
            }
        },
        restoreDraft : function(draft) {
            var $textarea;

            if (draft.id) {
                // The draft was editing a comment so find it and trigger the edit
                var $comment = this.$el.find('.comment[data-id=' + draft.id + ']');

                if ($comment.length) {
                    if (parseInt(draft.version, 10) >= parseInt($comment.attr('data-version'), 10)  ) {
                        // The comment hasn't been otherwise updated since saving the draft
                        $comment.find('> .content .edit').click();
                        $textarea = $comment.find(' .new-comment-form textarea');
                    } else {
                        // The comment has been edited since the draft, so delete the draft
                        this.deleteDraftComment(draft);
                    }
                }
            } else if (draft.parent) {
                // The draft was replying to a comment so find it and trigger the reply
                var $parent = this.$el.find('.comment[data-id=' + draft.parent.id + ']');

                if ($parent.length) {
                    $parent.find('> .content .reply').click();
                    $textarea = $parent.find('.replies .new-comment-form textarea');
                }
            } else {
                // The draft was a new general comment
                // Don't show it as collapsed because it has a draft
                var $form = this.$el.find('.general-comment-form').removeClass('collapsed');
                $textarea = $form.find('textarea');
            }

            if ($textarea) {
                $textarea
                    .val(draft.text)
                    .addClass('restored')
                    .attr("title",  stash_i18n('stash.web.comment.restored.draft.title', 'This comment has been restored from an automatically saved draft.'))
                    .trigger('input') //Fake an input to trigger initial sizing of textarea
                    .one('click keypress', function(){
                        $textarea
                            .removeClass('restored')
                            .removeAttr("title");
                    });
            } else {
                //we didn't find the right place to restore the draft, maybe it's on the next page of results.
                var self = this;

                eve.once('stash.feature.pullRequestActivity.dataLoaded', function(){
                    self.restoreDraft(draft);
                });
            }
        },
        closeCommentForm : function($form) {
            // don't close the general comment form, just empty it out. Clean up any restored draft attributes
            if ($form.is('.general-comment-form')) {
                $form.addClass('collapsed');
                $form.find('textarea')
                    .val('')
                    .trigger('input') //Fake an input to reset the size of the textarea
                    .removeClass('restored')
                    .removeAttr("title");
                //TODO DRY this up - MM
                this.deleteDraftComment(this.getDraftCommentFromForm($form));
                markupPreview.unbind($form);
            } else {
                CommentContainer.prototype.closeCommentForm.apply(this, arguments);
            }
        },
        getExtraCommentClasses : function() {
            return 'general-comment-activity';
        },
        onGeneralFormTextareaFocused : function(e) {
            var $form = $(e.target).closest('.general-comment-form');
            markupPreview.bindTo($form);
            $form.removeClass('collapsed');
        },
        destroy : function() {
            this.mentionableTextarea.destroy();
            delete this.mentionableTextarea;
        }
    });
});
