define('feature/comments/comment-container', [
    'jquery',
    'underscore',
    'eve',
    'util/client-storage',
    'model/page-state',
    'feature/comments/comment-model',
    'feature/comments/comment-collection',
    'feature/comments/comment-tips',
    'widget/markup-preview',
    'widget/aui/form',
    'widget/confirm-dialog'
], function (
    $,
    _,
    eve,
    clientStorage,
    pageState,
    Comment,
    Comments,
    commentTips,
    markupPreview,
    form,
    ConfirmDialog
    ) {

    var isMac = navigator.platform.indexOf('Mac') !== -1;

    /**
     * Backbone view. Requires:
     * * this.anchor || options.anchor,
     * * this.rootCommentListSelector || options.rootCommentListSelector
     */
    return Backbone.View.extend({
        initialize : function() {
            if (!this.collection) {
                this.collection = new Comments([], {
                    pullRequestId : pageState.getPullRequest() &&  pageState.getPullRequest().getId()
                });
            }
            this.anchor = this.anchor || this.options.anchor;
            this.rootCommentListSelector = this.rootCommentListSelector || this.options.rootCommentListSelector;
            this.context = this.options.context;
            this.initDeleteButtons();
            eve('stash.feature.comments.comment-container.added', this.$el);
        },
        events : {
            'submit form' : 'onFormSubmit',
            'click .cancel' : 'onCancelClicked',
            'click .reply' : 'onReplyClicked',
            'click .edit' : 'onEditClicked',
            'keydown textarea' : 'onTextareaKeydown'
        },
        initDeleteButtons : function(e) {
            this.createDeleteDialog().attachTo('.delete', null, this.el);
        },
        createDeleteDialog : function() {
            var self = this;
            var confirmDialog = new ConfirmDialog({
                id : "delete-repository-dialog",
                titleText: stash_i18n('stash.web.comment.delete.title', 'Delete comment'),
                titleClass : 'warning-header',
                panelContent : "<p>" + stash_i18n("stash.web.comment.delete.confirm", "Are you sure that you want to delete this comment?") + "</p>",
                submitText : stash_i18n('stash.web.button.delete', 'Delete'),
                submitToHref : false
            });

            confirmDialog.addConfirmListener(function(promise, $trigger, removeDialog) {
                removeDialog();
                self.deleteComment($trigger.closest('.comment'));
            });
            return confirmDialog;
        },
        onFormSubmit : function(e) {
            e.preventDefault();
            this.submitCommentForm($(e.target));
        },
        onCancelClicked : function(e) {
            e.preventDefault();
            this.cancelCommentForm($(e.target).closest('form'));
        },
        onReplyClicked : function(e) {
            e.preventDefault();
            this.openReplyForm($(e.target).closest('.comment'));
        },
        onEditClicked : function(e) {
            e.preventDefault();
            this.openEditForm($(e.target).closest('.comment'));
        },
        onTextareaKeydown : function(e) {
            if ((isMac ? e.metaKey : e.ctrlKey) && e.which === $.ui.keyCode.ENTER) { // Handle Ctrl+Enter/Cmd+Enter
                e.preventDefault();
                $(e.target).closest('form').submit();
            } else {
                // Unfortunately if you've written a draft, then focus in another comment form (e.g. general comment form) and press a keyboard shortcut (such as cmd+r)
                // it will treat it as a change, even though you haven't really typed anything, and it will overwrite your other draft. Bit of an edge case though.
                this.updateDraftComment(e.target);
            }
        },
        updateDraftComment : _.debounce(function(textArea) {
            var draft = this.getDraftCommentFromForm($(textArea).closest('form'));

            this.saveDraftComment(draft);
        }, 300),
        getDraftCommentFromForm: function($form){
            var draft = this.getCommentFormJSON($form);

            if (draft.anchor) {
                //commitRange adds a bunch of noise to the stored comment
                delete draft.anchor.commitRange;
            }

            // $.extend doesn't copy undefined properties. JSON.stringify throws them away, so this gives us a consistent object for equality checks,
            // regardless of whether it's retrieved from the form or from sessionStorage.
            return $.extend({}, draft);
        },
        saveDraftComment: function(draft) {
            if (this.context) {
                // We should be letting the context handle this but at the moment activity-comment-context
                // inherits from comment-container, so we need to handle these here too.
                this.context.saveDraftComment(draft);
            } else {
                clientStorage.setSessionItem(this.getDraftKey(), draft);
            }
        },
        deleteDraftComment : function(draft) {
            if (this.context) {
                this.context.deleteDraftComment(draft);
            } else {
                // We should only delete the draft when it's because we are cancelling/saving the currently drafted comment.
                // Otherwise if the user has multiple comments active (e.g. line comment + reply) they can clear a draft unexpectedly.
                var existingDraft = clientStorage.getSessionItem(this.getDraftKey());

                if (existingDraft) {
                    // Ignore the comment text in the equality check. (allows for quick press of cmd+enter before text has synchronised)
                    delete draft.text;
                    delete existingDraft.text;

                    if (_(draft).isEqual(existingDraft)) {
                        clientStorage.removeSessionItem(this.getDraftKey());
                    }
                }
            }
        },
        getDraftKey : function() {
            if (this.context) {
                return this.context.getDraftKey();
            } else {
                return clientStorage.buildKey(['draft-comment', this.anchor.getId()], 'user');
            }
        },
        getRootCommentList : function() {
            var $list = this.$(this.rootCommentListSelector);
            if (!$list.length) {
                $list = this.$el;
            }
            return $list;
        },
        render : function() {
            var $newEl = stash.feature.comments($.extend({
                comments : this.collection.toJSON(),
                currentUser : pageState.getCurrentUser(),
                tips : commentTips.tips
            }, this.anchor.toJSON()));

            this.$el.replaceWith($newEl);

            this.setElement($newEl[0]);
        },
        getCommentJSON : function($comment) {
            var parentId = $comment.parent().closest('.comment').data('id');
            return {
                id : $comment.attr('data-id'),
                version : $comment.attr('data-version'),
                text : $comment.find('> .content > .message').data('text'),
                anchor : this.anchor.toJSON(),
                parent : parentId ? { id : parentId } : undefined
            };
        },
        getCommentFormJSON : function($form) {
            var $commentRoot = $form.parent().is('.comment') ? $form.parent() : $form;
            var parentId = $commentRoot.parent().closest('.comment').data('id');
            return {
                id : $form.attr('data-id'),
                version : $form.attr('data-version'),
                text : $form.find('textarea').val(),
                anchor : this.anchor.toJSON(),
                parent : parentId ? { id : parentId } : undefined
            };
        },
        enableDeletion : function($comment) {
            var $actions = $comment.find('> .content > .actions');
            if (!$actions.children('.delete').length) {
                $actions.append('<a class="delete" href="#">' + stash_i18n('stash.web.comment.delete', 'Delete') + '</a><span class="separator">&middot;</span>');
            }
        },
        disableDeletion : function($comment) {
            $comment.find('> .content > .actions > .delete + .separator, > .content > .actions > .delete').remove();
        },
        renderContentUpdate : function($comment, commentJSON) {
            $comment.children('.content').replaceWith(stash.feature.commentContent($.extend({
                canReply: true,
                canEdit : true,
                canDelete : !$comment.children('.replies').children().length
            }, commentJSON)));
            $comment.attr('data-version', commentJSON.version).data('version', commentJSON.version);
            eve('stash.feature.comments.comment.edited', $comment, commentJSON, this.context);
        },
        insertCommentIntoList : function($comment, $commentList, commentJSON) {
            var $insertBefore = $commentList.children('.comment:first');
            //TODO HACK: sort by ID is implied sort by createdDate. is that even what we want?
            while ($insertBefore.length) {
                if (parseInt($insertBefore.data('id'), 10) > commentJSON.id) {
                    break;
                }
                $insertBefore = $insertBefore.next('.comment');
            }

            $insertBefore = $insertBefore.length ?
                                $insertBefore :
                                $commentList.children('.comment-form-container:last');

            if ($insertBefore.length) {
                $comment.insertBefore($insertBefore);
            } else {
                $commentList.append($comment);
            }
        },
        renderComment : function(commentJSON, parentId, isContentUpdate) {

            var $comment;

            if (isContentUpdate && ($comment = $('.comment[data-id="' + commentJSON.id + '"]')).length) {
                return this.renderContentUpdate($comment, commentJSON);
            }

            commentJSON = $.extend({
                isNew : true
            }, commentJSON);

            var $parent = parentId && this.$('[data-id=' + parentId + ']');

            if ($parent) { // disable deleting the parent now that it has replies
                this.disableDeletion($parent);
            }

            var $insertUnder = ($parent ? // a reply
                 $parent.children('.replies') :
                 this.getRootCommentList());

            $comment = $(stash.feature.comment($.extend({
                currentUser: pageState.getCurrentUser(),
                numOfAncestors: $insertUnder.parents('.comment').length,
                extraClasses: this.getExtraCommentClasses()
            }, commentJSON)));

            this.insertCommentIntoList($comment, $insertUnder, commentJSON);

            // Should match the timing of the target-fade-animation in comments.less
            // We remove the clsas because Chrome bugs out and replays the animation when switching between
            // fixed and normal modes on the diff page.
            var targetFadeAnimationTime = 5000;
            setTimeout(_.bind($comment.removeClass, $comment, 'new'), targetFadeAnimationTime);

            $('html, body').animate({
                scrollTop: Math.max(0, $comment.offset().top - $(window).height() / 2)
            });

            eve('stash.feature.comments.comment.added', $comment, commentJSON, this.context);
            $comment.hide().fadeIn('slow');
        },
        getExtraCommentClasses : function() {
            return '';  // Can be overridden for different contexts
        },
        showErrorMessage : function(model, error) {
            var $form = this;
            var $error = $form.find('.error');
            if (!$error.length) {
                $error = $('<div class="error"></div>');
                $form.find('.buttons').before($error);
            }
            $error.text(error);
        },
        cancelCommentForm : function($form) {
            this.closeCommentForm($form);
        },
        submitCommentForm : function($form) {
            if (form.isSubmissionPrevented($form)) {
                return;
            }

            var self = this;
            var spinner = $form.find('.spinner');

            var commentJSON = this.getCommentFormJSON($form);
            var isEdit = commentJSON.id != null;
            var isKnown = isEdit && this.collection.get(commentJSON.id);
            var parentId = commentJSON.parent && commentJSON.parent.id;
            var comment = isKnown ? this.collection.get(commentJSON.id) : new Comment();

            comment.on('error', this.showErrorMessage, $form);
            if (comment.set($.extend(commentJSON, { avatarSize: stash.widget.avatarSizeInPx({ size: 'medium' }) }))) {
                if (!isKnown) {
                    this.collection.push(comment);
                }
                // need to override the positioning - spin.js can't handle the absolute positioning of the container.
                spinner.spin('small', { top: 0, left: 0 });
                form.preventSubmission($form);
                comment.save().done(function(commentResp) {
                    // hack to avoid "future" comments
                    commentResp.createdDate = Math.min(commentResp.createdDate, new Date().getTime());
                    commentResp.updatedDate = Math.min(commentResp.updatedDate, new Date().getTime());

                    self.closeCommentForm($form, { doNotDestroy : true });
                    self.renderComment(commentResp, parentId, isEdit);
                }).fail(function() {
                    if (!isKnown && !isEdit) { // it was a totally new comment. remove it from our models.
                        self.collection.remove(commentJSON.id);
                    }
                }).always(function() {
                    spinner.spinStop();
                    form.allowSubmission($form);
                });
            }
            comment.off('error', this.showErrorMessage);
        },
        deleteComment : function($comment) {
            var commentJSON = this.getCommentJSON($comment);
            var comment;
            if (this.collection.get(commentJSON)) {
                comment = this.collection.get(commentJSON.id);
            } else {
                comment = new Comment(commentJSON);
                this.collection.push(comment);
            }
            var $delete = $comment.find('> .content .delete'),
                deleteDims = { h : $delete.height(), w : $delete.width() };
            $delete.height(deleteDims.h).width(deleteDims.w).css('vertical-align', 'middle').empty().spin('small');

            var self = this;
            comment.destroy({ wait: true }).always(function() {
                $delete.spinStop();
            }).done(function() {
                var $parent = $comment.parent().closest('.comment');
                if ($parent.length) {
                    // has siblings if there 1) are sibling comments, or 2) there is a form being edited
                    var hasSiblings = $comment.siblings('.comment').length || $comment.siblings('.comment-form-container').find('textarea').val();
                    if (!hasSiblings) { // this was the only child. reenable deletion on its parent
                        self.enableDeletion($parent);
                    }
                }
                $comment.fadeOut(function () {
                    $comment.remove();
                    self.onCommentDeleted();
                    eve("stash.feature.comments.comment.deleted", commentJSON, self.context);
                });

            }).fail(function() { // revert to Delete
                $delete.css('height', '').css('width', '').css('vertical-align', '').text(stash_i18n('stash.web.comment.delete', 'Delete'));
            });
        },
        onCommentDeleted : function() {
            // Can be overridden
        },
        openEditForm : function($comment) {
            var commentJSON = this.getCommentJSON($comment);
            var $form = $(stash.feature.commentForm($.extend({ tips : commentTips.tips, currentUser: pageState.getCurrentUser() }, commentJSON)));
            var $originalContent = $comment.find('> .user-avatar, > .content');
            $originalContent.replaceWith($form);
            $comment.addClass('comment-form-container');
            $form.data('originalContent', $originalContent);
            $form.find('textarea.expanding').expandingTextarea();
            // IE scrolls the diff pane left to the start of the textarea when calling focus() immediately
            _.defer(function(){ $form.find('textarea').focus(); });
            eve('stash.feature.comments.comment-form.displayed', $form);
            markupPreview.bindTo($form);
        },
        openReplyForm : function($replyToComment) {
            var $replies = $replyToComment.children('.replies');
            this.openCommentForm($replies, { location: 'top' });
        },
        openNewCommentForm : function() {
            this.openCommentForm(this.getRootCommentList(), { location: 'bottom' });
        },
        openCommentForm : function($commentList, options) {
            var attachmentMethod = options && options.location === 'top' ? 'prependTo' : 'appendTo';
            // check if there is a form open that is not an edit form.
            var $formContainer = $commentList.children('.comment-form-container').not('.comment');
            if (!$formContainer.length) {
                $formContainer = $(stash.feature.commentFormListItem({ tips : commentTips.tips, currentUser: pageState.getCurrentUser() }))[attachmentMethod]($commentList);
            }
            $formContainer.find('textarea.expanding').expandingTextarea();
            // IE scrolls the diff pane left to the start of the textarea when calling focus() immediately
            _.defer(function(){ $formContainer.find('textarea').focus(); });
            var $form = $formContainer.find('form');
            markupPreview.bindTo($form);
            eve('stash.feature.comments.comment-form.displayed', $form);
        },
        closeCommentForm : function($form) {
            //TODO DRY this up - MM
            this.deleteDraftComment(this.getDraftCommentFromForm($form));

            markupPreview.unbind($form);

            var $originalContent = $form.data('originalContent');
            var $li = $form.parent();

            if ($originalContent) { // edits - the form is inside a living comment, so revert back to that comment's content.
                $li.removeClass('comment-form-container');

                $form.replaceWith($originalContent);
            } else { // it's a new comment - remove it entirely.
                $li.remove();
            }
            eve('stash.feature.comments.comment-form.closed', $form);
        }
    });
});
