define('feature/comments/diff-comment-context', [
    'jquery',
    'underscore',
    'eve',
    'backbone',
    'util/performance',
    'util/client-storage',
    'model/page-state',
    'widget/mentionable-textarea',
    'feature/comments/diff-comment-container',
    'feature/comments/anchors',
    'feature/comments/comment-tips'
], function (
    $,
    _,
    eve,
    Backbone,
    performance,
    clientStorage,
    pageState,
    MentionableTextarea,
    DiffCommentContainer,
    anchors,
    commentTips
) {
    var LineAnchor = anchors.LineAnchor;

    return Backbone.View.extend({
        initialize : function() {
            this._containers = {
            };
            this.checkForNewContainers();

            var self = this;

            eve.on('window.resize.debounced', this._resizeHandler = function() {
                self.fitToWidth();
            });
            eve.on('stash.feature.changeset.difftree.collapseAnimationFinished', this._resizeHandler);

            eve.on('stash.feature.comments.comment.added', this._commentAddedHandler = function(commentJson, context) {
                if (context === self && self.$el.find('.comment').length === 1) {
                    eve('stash.feature.comments.first.comment.added');
                }
            });

            var draft = clientStorage.getSessionItem(this.getDraftKey());
            if (draft && draft.text) {
                // When refactoring, do not wrap this in a _.defer, as it will cause the diffview.dataLoaded event to be bound after it fires
                this.restoreDraft(draft);
            }
            this.mentionableTextarea = new MentionableTextarea({
                selector: '.comment-form-container textarea',
                $container: this.$el
            });

        },
        events : {
            'click .line:not(.conflict-marker) .add-comment-trigger' : 'addCommentClicked'
        },
        registerContainer : function(containerEl, anchor, opt_containerName) {
            opt_containerName = opt_containerName || anchor.getId();
            if (!_.has(this._containers, opt_containerName)) {
                this._containers[opt_containerName] = new DiffCommentContainer({
                    name : opt_containerName,
                    context : this,
                    el : containerEl,
                    anchor : anchor
                });
                var self = this;
                var setMaxWidth = function() {
                    self._containers[opt_containerName].setPanelMaxWidthPx(self._width);
                };
                if (!this._width) {
                    performance.queueDOMRead(function() {
                        self._width = self._width || self.$el.width();
                        setMaxWidth();
                    });
                } else {
                    setMaxWidth();
                }
            }
        },
        checkForNewContainers : function() {
            var self = this;
            _.forEach(this.$('.line + .comment-box'), function(commentContainer) {
                self.registerContainer(commentContainer, self.getLineAnchor($(commentContainer).prev()));
            });
        },
        addCommentClicked : function(e) {
            e.preventDefault();

            var $line = $(e.target).closest('.line');
            var container = this.addCommentContainerForLine($line);
            container.openNewCommentForm();
        },
        addCommentContainerForLine : function($line) {
            var lineAnchor = this.getLineAnchor($line);
            var containerName = lineAnchor.getId();
            if (_.has(this._containers, containerName)) {
                // we already know about this container and it's being handled.
                return this._containers[containerName];
            }

            var $commentContainer = $line.next();
            if (!$commentContainer.is('.comment-box')) { // comment container DOM needs to be created
                $commentContainer = $(stash.feature.comments($.extend({
                    tips : commentTips.tips,
                    currentUser : pageState.getCurrentUser(),
                    extraClasses: 'comment-box'
                }, lineAnchor.toJSON())));
                $commentContainer.insertAfter($line);
            }
            this.registerContainer($commentContainer[0], lineAnchor, containerName);

            return this._containers[containerName];
        },
        getLineAnchor : function($line) {
            return new LineAnchor(this.options.anchor, $line.data('line-type'), $line.data('line'));
        },
        fitToWidth : function() {
            this._width = this.$el.width();
            for(var containerName in this._containers) {
                if (this._containers.hasOwnProperty(containerName)) {
                    var container = this._containers[containerName];
                    container.setPanelMaxWidthPx(this._width);
                }
            }
        },
        saveDraftComment: function(draft) {
            clientStorage.setSessionItem(this.getDraftKey(), draft);
        },
        deleteDraftComment : function(draft) {
            // We should only delete the draft when it's because we are cancelling/saving the currently drafted comment.
            // Otherwise if the user has multiple comments active (e.g. line comment + reply) they can clear a draft unexpectedly.
            var existingDraft = clientStorage.getSessionItem(this.getDraftKey());

            if (existingDraft) {
                // Ignore the comment text in the equality check. (allows for quick press of cmd+enter before text has synchronised)
                delete draft.text;
                delete existingDraft.text;

                if (_(draft).isEqual(existingDraft)) {
                    clientStorage.removeSessionItem(this.getDraftKey());
                }
            }
        },
        getDraftKey: function() {
            return clientStorage.buildKey(['draft-comment', this.options.anchor.getId()], 'user');
        },
        restoreDraft : function(draft) {
            var self = this;

            if(draft.anchor && draft.anchor.line) {
                // This is a line comment
                var diffLoadedHandler = function(start, limit, data){
                    if (draft.anchor.path === (data.destination || data.source).components.join("/")) {
                        // This is the diff/droids we're looking for
                        eve.off("stash.feature.diffview.dataLoaded", diffLoadedHandler);
                    } else {
                        return;
                    }

                    var $textarea;

                    if (draft.id) {
                        // The draft was editing a comment so find it and trigger the edit
                        var $comment = self.$el.find('.comment[data-id=' + draft.id + ']');

                        if ($comment.length) {
                            if (parseInt(draft.version, 10) >= parseInt($comment.attr('data-version'), 10)  ) {
                                // The comment hasn't been otherwise updated since saving the draft
                                $comment.find('> .content .edit').click();
                                $textarea = $comment.find(' .new-comment-form textarea');
                            } else {
                                // The comment has been edited since the draft, so delete the draft
                                self.deleteDraftComment(draft);
                            }
                        }
                    } else if (draft.parent) {
                        // The draft was replying to a comment so find it and trigger the reply
                        var $parent = self.$el.find('.comment[data-id=' + draft.parent.id + ']');

                        if ($parent.length) {
                            $parent.find('> .content .reply').click();
                            $textarea = $parent.find('.replies .new-comment-form textarea');
                        }
                    } else {
                        // The draft was a new line comment. Only trigger if the comment container is ".commentable", i.e. allows new comments;
                        var $line = self.$el.filter('.commentable').find('.line[data-line="' + draft.anchor.line + '"][data-line-type="' + draft.anchor.lineType + '"]');

                        if ($line.length) {
                            $line.find('.add-comment-trigger').click();
                            $textarea = $('#comments-' + draft.anchor.lineType + "-" + draft.anchor.line).find('.new-comment-form textarea');
                        }
                    }

                    if ($textarea) {
                        $textarea
                            .val(draft.text)
                            .addClass('restored')
                            .attr("title",  stash_i18n('stash.web.comment.restored.draft.title', 'This comment has been restored from an automatically saved draft.'))
                            .trigger('input') //Fake an input to trigger initial sizing of textarea
                            .one('click keypress', function(){
                                $textarea
                                    .removeClass('restored')
                                    .removeAttr("title");
                            });
                    } else {
                        // TODO: Currently we don't page the diff, so we can guarantee that if we didn't find the comment the first time, it doesn't exist.
                        // TODO: This will change when (if) we start paging the diff.
                    }
                };
                eve.on("stash.feature.diffview.dataLoaded", diffLoadedHandler);
            } else {
                //TODO: Handle file comments when we implement them.
            }
        },
        destroy : function(opt_container) {
            if (opt_container) {
                opt_container.remove();
                delete this._containers[opt_container.options.name];

                if (!this.$el.has('.comment').length) {
                    eve('stash.feature.comments.last.comment.deleted');
                }
            } else {
                _.invoke(this._containers, 'remove');
                this._containers = null;

                if (this._resizeHandler) {
                    eve.off('window.resize.debounced', this._resizeHandler);
                    eve.off('stash.feature.changeset.difftree.collapseAnimationFinished', this._resizeHandler);
                    eve.off('stash.feature.comments.comment.added', this._commentAddedHandler);
                    delete this._resizeHandler;
                    delete this._commentAddedHandler;
                }
                this.mentionableTextarea.destroy();
                delete this.mentionableTextarea;
            }
        }
    });
});
