define('feature/file-content/diff-image', ['jquery', 'underscore', 'eve', 'exports'], function ($, _, eve, exports) {

    // fdSlider triggers a change event in js when the polyfill is used.
    // something about how the event is fired means jQuery doesn't pick it up, so handlers attached through Query don't fire
    // I've added this function as a guard against memory leaks when assigning listeners natively.
    function addChangeListener(el, fn) {
        el.onchange = fn;
    }

    function removeChangeListeners(el) {
        el.onchange = null;
    }

    function calculateImageNaturalSize($image) {

        if ($image.data('natural-size')) {
            return $.Deferred().resolve($image.data('natural-size'));
        }

        // user the naturalWidth and naturalHeight properties if the browser supports it
        if ($image[0].naturalWidth) {
            var size = {width: $image[0].naturalWidth, height: $image[0].naturalHeight};
            $image.data('natural-size', size);
            return $.Deferred().resolve(size);
        }

        // otherwise create an image object in memory, wait till it's loaded and get its dimensions (IE8)
        var newImg = new Image(),
            promise = $.Deferred(),
            returned = false,
            onComplete = function() {
                if (!returned) {
                    var size = {width: newImg.width, height: newImg.height};
                    $image.data('natural-size', size);
                    promise.resolve(size);
                    returned = true;
                }
            };

        newImg.onload = onComplete;
        newImg.src = $image[0].src;
        if (newImg.complete) {
            onComplete();
        }
        return promise;
    }


    function ImageDiffer($container) {
        this._$container = $container;
    }

    ImageDiffer.prototype.init = function() {
        var self = this;

        var oldSize;
        var newSize;

        //add the diff mode toolbar buttons, split and blend buttons disabled by default
        this._$toggle = $(stash.feature.fileContent.imageDiffToggle()).appendTo(this._$container.find('.image-diff-toolbar'));
        this._$toggle.on('click', function(e) { e.preventDefault(); });

        var $imgs = this._$imgs = this._$container.find('img');

        this._$sinceImg = $imgs.eq(0);
        this._$untilImg = $imgs.eq(1);

        this._$untilRevision = this._$untilImg.parent();
        this._$sinceRevision = this._$sinceImg.parent();
        this._$revisions = this._$untilRevision.add(this._$sinceRevision);

        this._$untilRevisionLabel = self._$untilRevision.find('h5');
        this._$sinceRevisionLabel = self._$sinceRevision.find('h5');

        $imgs.imagesLoaded().always(function() {

            var sincePromise = calculateImageNaturalSize(self._$sinceImg).done(function(size) { oldSize = size; });
            var untilPromise = calculateImageNaturalSize(self._$untilImg).done(function(size) { newSize = size; });

            $.when(sincePromise, untilPromise).done(function() {
                var sameDimensions = (oldSize.width === newSize.width) && (oldSize.height === newSize.height);

                var enableExtraModes = sameDimensions && oldSize.width > 0;

                self._initToolbar(enableExtraModes);
            });
        });
    };

    ImageDiffer.prototype._initToolbar = function (enableExtraModes) {
        var self = this;

        var $diffTwoUpModeButton = this._$toggle.find('.image-diff-two-up');
        var $diffBlendModeButton = this._$toggle.find('.image-diff-blend');
        var $diffSplitModeButton = this._$toggle.find('.image-diff-split');

        // enable the split and blend buttons if we're supporting that mode.
        $diffSplitModeButton.attr('aria-disabled', !enableExtraModes);
        if (enableExtraModes) {
            $diffSplitModeButton.attr('title', stash_i18n('stash.web.sourceview.button.diff.image.split.tooltip', 'Compare the image changes by moving your mouse cursor left or right over the image'));
        }

        $diffBlendModeButton.attr('aria-disabled', !enableExtraModes);
        if (enableExtraModes) {
            $diffBlendModeButton.attr('title', stash_i18n('stash.web.sourceview.button.diff.image.blend.tooltip', 'Compare the image changes by fading in the differences'));
        }

        var $modeListButtons = this._$toggle.find('.aui-button');
        var oldMode = self._oldMode = 'two-up';
        var makeButtonClick = function(newMode) {
            return function(e) {
                var $this = $(this);
                if ($this.attr('aria-disabled')!=='true' && $this.attr('aria-pressed')!=='true') {
                    $modeListButtons.attr('aria-pressed', 'false');
                    $this.attr('aria-pressed', 'true');

                    self._onDiffModeChanged(newMode, oldMode);

                    oldMode = self._oldMode = newMode;
                }
                e.preventDefault();
            };
        };
        $diffTwoUpModeButton.click(makeButtonClick('two-up'));
        $diffBlendModeButton.click(makeButtonClick('blend'));
        $diffSplitModeButton.click(makeButtonClick('split'));

        $diffTwoUpModeButton.add($diffBlendModeButton).add($diffSplitModeButton)
            .tooltip({
                gravity: 's'
            });

        this._toolbarInitialized = true;
    };

    ImageDiffer.prototype.destroy = function() {
        this._cleanOldMode(this._oldMode);
        if (this._toolbarInitialized) {
            this._destroyToolbar();
            this._toolbarInitialized = false;
        }
    };

    ImageDiffer.prototype._destroyToolbar = function() {
        this._$toggle.remove();
    };

    ImageDiffer.prototype._cleanOldMode = function(oldMode) {
        switch(oldMode) {
            case 'two-up':
                this._cleanTwoUpDiff();
                break;
            case 'blend':
                this._cleanBlendDiff();
                break;
            case 'split':
                this._cleanSplitDiff();
                break;
        }
    };

    ImageDiffer.prototype._onDiffModeChanged = function(newMode, oldMode) {

        this._$container.removeClass('two-up blend split').addClass(newMode);

        this._cleanOldMode(oldMode);

        switch (newMode) {
            case 'two-up':
                this._setupTwoUpDiff();
                break;
            case 'blend':
                this._setupBlendDiff();
                break;
            case 'split':
                this._setupSplitDiff();
                break;
        }

        eve('stash.feature.fileContent.diffView.imageDiffModeChanged', this, newMode, oldMode);
    };

    ImageDiffer.prototype._cleanTwoUpDiff = $.noop;
    ImageDiffer.prototype._setupTwoUpDiff = $.noop;

    ImageDiffer.prototype._cleanBlendDiff = function() {
        var $opacitySliderContainer = this._$container.children('.opacity-slider-container');
        var $opacitySlider = $opacitySliderContainer.find('input');

        $opacitySliderContainer.remove();
        this._$untilImg.css('opacity', '');

        removeChangeListeners($opacitySlider[0]);
    };

    ImageDiffer.prototype._setupBlendDiff = function() {
        var self = this;

        // add in the slider
        var $opacitySlider = $('<input type="range" min="0" max="1" step="0.01" value="1" />');

        this._$container.children('.image-diff-toolbar').after(
            $('<div class="opacity-slider-container" />').append($opacitySlider)
        );

        fdSlider.onDomReady();

        // set opacity when the slider changes
        addChangeListener($opacitySlider[0], function () {
            self._$untilImg.css('opacity', parseFloat($(this).val()));
        });
    };

    ImageDiffer.prototype._cleanSplitDiff = function() {

        this._$imgs.css('width', '').css('height', '');
        this._$revisions.css('width', '');
        this._$untilRevision.css('margin-left', '');
        this._$untilRevisionLabel.add(this._$sinceRevisionLabel).css('margin-left', '').css('margin-right', '');

        this._$untilImg.unwrap();
        this._$untilRevisionImageContainer.remove(); // cleanup events
        delete this._$untilRevisionImageContainer;

        this._$revisions.unwrap();
        this._$splitContainer.remove(); // cleanup events
        delete this._$splitContainer;

        if (this._onResize) {
            eve.off('window.resize.debounced', this._onResize);
            delete this._onResize;
        }
    };

    function getMaxWidthToFit($el, parentWidth) {
        var marginBorderPadding = $el.data('margin_border_padding');
        if (marginBorderPadding == null) {
            marginBorderPadding = $el.outerWidth(true) - $el.width();
            $el.data('margin_border_padding', marginBorderPadding);
        }

        return parentWidth - marginBorderPadding;
    }

    ImageDiffer.prototype._setSplitDiffElementProperties = function($untilRevisionImageContainer, $splitContainer, naturalImageSize) {

        // determine the chain of element widths that will allow us to fit our imgs within the content frame:
            // max width of .binary-container
        var maxBinaryContainerWidth = getMaxWidthToFit(this._$container, this._$container.parent().width());
            // max width of .split-container
        var maxSplitContainerWidth = getMaxWidthToFit($splitContainer, maxBinaryContainerWidth);
            // max width of each .binary
        var maxRevisionWidth = getMaxWidthToFit(this._$sinceRevision, maxSplitContainerWidth);
            // max width of each image
        var maxImgWidth = getMaxWidthToFit(this._$sinceImg, maxRevisionWidth);

        // set the imgs to their natural width explicitly, or else to the max width if they won't fit.
        var imageWidth = Math.min(naturalImageSize.width, maxImgWidth);
        this._$imgs.width(imageWidth);
        // set a proportional height
        this._$imgs.height(Math.floor((imageWidth / naturalImageSize.width) * naturalImageSize.height));

        // set the image container height to perfectly fit the images (they have a border, so we can't just use the same value).
        $untilRevisionImageContainer.height(this._$imgs.outerHeight(true));


        // now that we have the width of the images, work back up so we can set an explicit width on .binary
        var revisionWidth = imageWidth + (maxRevisionWidth - maxImgWidth);
        this._$revisions.width(revisionWidth);
        // shift the entire .until-revision container to the left by the same distance as the width of the image, so that the two .binary containers overlap
        this._$untilRevision.css('margin-left', -revisionWidth);
    };

    ImageDiffer.prototype._setupSplitDiff = function() {
        var self = this;

        return calculateImageNaturalSize(this._$imgs).done(function(naturalImageSize) {

            self._$untilImg.wrap('<div class="image-container" />');
            self._$revisions.wrapAll('<div class="split-container" />');

            self._$untilRevisionImageContainer = self._$untilImg.parent();
            self._$splitContainer = self._$container.find('.split-container');


            var maxImageContainerWidth;
            self._onResize = function() {
                self._setSplitDiffElementProperties(self._$untilRevisionImageContainer, self._$splitContainer, naturalImageSize);

                maxImageContainerWidth = self._$sinceImg.outerWidth(true);
            };
            self._onResize();
            eve.on('window.resize.debounced', self._onResize);

            if(naturalImageSize.width < 50) {
                // If the image width is < 50px, that offset based on image width is not going to be sufficient to prevent overlap, so fix it at -50px
                self._$untilRevisionLabel.css('margin-left', '-50px');
                self._$sinceRevisionLabel.css('margin-right', '-50px');
            } else if(naturalImageSize.width < 100) {
                // If the image width is < 100px, offset the since and until (old/new) labels by the same px distance using negative margins so they sit outside of the .binary container
                self._$untilRevisionLabel.css('margin-left', -naturalImageSize.width);
                self._$sinceRevisionLabel.css('margin-right', -naturalImageSize.width);
            }

            var offset;
            self._$splitContainer.on('mouseenter', function() {
                offset = self._$untilRevision.offset();
                self._$revisions.on('mousemove', onMove);
            });

            self._$splitContainer.on('mouseleave', function() {
                self._$revisions.off('mousemove', onMove);
            });

            var onMove = function(e)  {
                self._$untilRevisionImageContainer.css('width', Math.min(e.pageX - offset.left, maxImageContainerWidth));
            };
        });
    };

    var differs = [];
    function handleBinary(data) {
        if (data.sourceType === data.destinationType && data.sourceType === 'image') {
            var differ = new ImageDiffer(data.$container);
            differ.init();
            differs.push(differ);
        }
    }
    exports.init = function () {
        eve.on('stash.feature.diffview.onBinary', handleBinary);
    };
    exports.reset = function() {
        eve.off('stash.feature.diffview.onBinary', handleBinary);
        _.invoke(differs, 'destroy');
        differs = [];
    };
});
