define('feature/file-content/diff-view',
[
    'jquery',
    'underscore',
    'eve',
    'aui',
    'util/ajax',
    'util/navbuilder',
    'model/page-state',
    'util/html',
    'util/error',
    'model/path',
    'widget/paged-scrollable',
    'feature/comments',
    'feature/file-content/binary-view',
    'feature/ediff'
], function (
    $,
    _,
    eve,
    AJS,
    ajax,
    navBuilder,
    pageState,
    htmlUtil,
    errorUtil,
    Path,
    PagedScrollable,
    comments,
    binaryView,
    ediff
) {

    var pageSize = 10000,
        maxAnimatedLength = 2000000,
        lineNodeStream = htmlUtil.lineNodeStream,
        mergeStreams = htmlUtil.mergeStreams,
        generateEdiffRegions = ediff.diff,
        DiffRegionNodeStream = ediff.DiffRegionNodeStream,
        tokenizeString = ediff.tokenizeString,
        isEdiffDisabled = $.browser.msie && parseInt($.browser.version, 10) < 9,
        ediffTokenLimit = 750; // 10 lines * 75 tokens (roughly indicative of a long code line) per line. Open to tweaking

    var tips = comments.tips;

    function getData(restUrl) {
        return ajax.rest({
            url: restUrl,
            statusCode : {
                '404' : function(xhr, testStatus, errorThrown, data, fallbackError) {

                    var error = data && data.errors && data.errors.length && data.errors[0];

                    if (errorUtil.isErrorEntityWithinRepository(error)) {
                        return false; // don't handle this globally.
                    }
                }
            }
        });
    }

    function DiffView(containerSelector, scrollPaneSelector, options) {

        PagedScrollable.call(this, scrollPaneSelector, $.extend({
            pageSize: pageSize,
            dataLoadedEvent: 'stash.feature.diffview.dataLoaded'
        }, options));

        this.showComments = this.options.commentMode === comments.commentMode.READ ||
                            this.options.commentMode === comments.commentMode.CREATE_NEW ||
                            this.options.commentMode === comments.commentMode.REPLY_ONLY;
        this.allowCommenting = this.options.commentMode === comments.commentMode.CREATE_NEW;

        this.showingBinary = false;

        this.$container = $(containerSelector).toggleClass('commentable', this.allowCommenting);
        this.$fromLines = this.$container.children('.from-lines');
        this.$toLines = this.$container.children('.to-lines');
        this.$diffContent = this.$container.find('.source-container');
    }
    $.extend(DiffView.prototype, PagedScrollable.prototype);

    DiffView.prototype.init = function(fileChange, options) {
        this.reset();

        this.fileChange = fileChange;
        this.options = $.extend({}, this.options, options);

        this.preloaded = !!fileChange.getDiff();

        // if we're going to need the spinner, initialize it
        if (!this.$spinnerTarget && !this.preloaded) {
            this.$spinnerTarget = $('<div id="file-content-spinner" />');
            this.$container.after(this.$spinnerTarget);
        }

        var self = this;
        this.onRequestedRevisionData = function() {
            self.reset();
            self.$container.removeClass('fully-loaded');
            if (!self.preloaded) {
                self.$spinnerTarget.show().spin("large");
            }
        };
        eve.on('stash.layout.branch.requestedRevisionData', this.onRequestedRevisionData);

        if (this.showComments) {
            comments.bindContext(this.$container, new comments.DiffAnchor(fileChange), {
                allowCommenting : this.allowCommenting
            });
        }

        return PagedScrollable.prototype.init.call(this);
    };

    DiffView.prototype.reset = function() {
        this._restoreTable();

        this.$fromLines.empty();
        this.$toLines.empty();
        this.$diffContent.empty();

        if(this.onRequestedRevisionData) {
            eve.unbind('stash.layout.branch.requestedRevisionData', this.onRequestedRevisionData);
        }

        if (this.showComments) {
            comments.unbindContext(this.$container);
        }

        PagedScrollable.prototype.reset.call(this);
    };

    DiffView.prototype.destroy = function() {
        this.reset();
        if (this.$spinnerTarget) {
            this.$spinnerTarget.remove();
            this.$spinnerTarget = null;
        }
    };

    //@Override
    DiffView.prototype.requestData = function (start, limit) {
        var self = this;

        if (this.preloaded) {
            return $.Deferred().resolve($.extend({
                isLastPage : true,
                size : pageSize,
                lineComments : this.options.lineComments,
                fileComments : this.options.fileComments
            }, this.fileChange.getDiff()));
        }

        this.$spinnerTarget.spin('large');

        var xhr = getData(navBuilder.currentRepo()
                                    .diff(this.fileChange)
                                    .withParams({avatarSize: stash.widget.avatarSizeInPx({ size: 'medium' }), markup: true})
                                    .build()),
            pipedXhr = xhr.pipe(function (data) {
                if (data.errors && data.errors.length) {
                    return $.Deferred().rejectWith(this, [this, null, null, data]);
                } else if ($.isArray(data.diffs) && data.diffs.length) {
                    // Data is returned as an array due to the REST resource supporting multiple diffs
                    // The first will not necessarily be the one we want. We want the first with the destination
                    // set to the selected file. This avoids choosing the wrong diff where a file has been modified
                    // and copied in the same changeset, for example.
                    data = _.find(data.diffs, function(diff) {
                        if (diff.destination) {
                            return diff.destination.toString === self.fileChange.getPath().toString();
                        } else if (self.fileChange.getSrcPath()) {
                            return diff.source.toString === self.fileChange.getSrcPath().toString();
                        }
                        return false;
                    }) || data.diffs[0]; //Or the first diff if none were found (this shouldn't happen)
                }

                //TODO: PagedScrollable probably should be agnostic to the variables used for paging.
                // for now, hacking in the interface it recognizes
                //assume always full page received
                data.isLastPage = true;
                data.size = pageSize;

                return data;
            }).
            always(function (xhr, text, error, data) {
                self.$spinnerTarget.spinStop();
            });

        // return the original xhr, but with the piped done|fail|notify methods.
        return pipedXhr.promise(xhr);
    };

    DiffView.prototype.handleErrors = function(errors) {
        if (errors) {
            this.$diffContent.prepend(_.reduce(errors, function(seed, error) {
                return error && error.message ?
                       seed + '<div class="aui-message error">' + AJS.escapeHtml(error.message) + '</div>' :
                       seed;
            }, ''));
        }
    };

    DiffView.prototype.onBinary = function (data) {
        var self = this;
        var $container = this.$container.find(".binary-container").empty();


        binaryView.renderBinaryDiff.call(this, $container, data.source, data.destination, this.fileChange.getCommitRange());
        if (!this.preloaded) {
            this.$spinnerTarget.spinStop();
        }
    };

    //@Override
    DiffView.prototype.onDataLoaded = function (start, limit, data) {
        if (binaryView.shouldRenderBinary(data)) {
            if (!this.showingBinary) {
                this.showingBinary = true;
                this.$container
                        .find(".source-container").hide()
                        .end()
                        .find(".binary-container").show();
            }
            this.onBinary(data);
        } else {
            if (this.showingBinary) {
                this.showingBinary = false;
                this.$container
                        .find(".source-container").show()
                        .end()
                        .find(".binary-container").hide().empty();
            }
            PagedScrollable.prototype.onDataLoaded.call(this, start, limit, data);
        }

        if (!this.preloaded) {
            this.$spinnerTarget.spinStop();

            if (data.isLastPage || data.binary) {
                    this.$container.addClass('fully-loaded');
                    this.$spinnerTarget.hide();
            }
        }
    };

    DiffView.prototype._replaceTable = function(replaceWith) {
        if (!this._$replacer) {
            this._$replacer = $(replaceWith);
            this.$container.replaceWith(this._$replacer);
        } else {
            var $oldReplacer = this._$replacer;
            this._$replacer = $(replaceWith);
            $oldReplacer.replaceWith(this._$replacer);
        }
    };
    DiffView.prototype._restoreTable = function() {
        if (this._$replacer) {
            this._$replacer.replaceWith(this.$container);
            this._$replacer = null;
        }
    };

    DiffView.prototype.attachNewContent = function (data, attachmentMethod) {
        if (this.fileChange.getConflict()) {
            this.$container.before(renderConflictMessage(this.fileChange.getConflict()));
        }

        if (data && data.errors) {
            this._replaceTable(renderErrors(data));
            return;
        }

        if (!data || !data.hunks || !data.hunks.length) {
            this.$spinnerTarget.remove();
            this._replaceTable(renderEmptyDiff(data));
            return;
        }

        if (data.hunks[data.hunks.length - 1].truncated) {
            this._replaceTable(renderTooLargeDiff(
                this.fileChange.getCommitRange(),
                new Path(data.destination)
            ));
            return;
        }

        var content = renderDiffContent(data, this.showComments, this.allowCommenting);
        if (!this.preloaded && content.length < maxAnimatedLength) {
            this.$diffContent.hide()[attachmentMethod](content).fadeIn('fast');
        } else {
            this.$diffContent[attachmentMethod](content);
        }

        if (this.showComments) {
            comments.updateContext(this.$container);
        }
    };

    DiffView.prototype.selectLine = function(lineNumber, initialPageLoad) { /*TODO: NOOP for now.*/ };

    var makeLineWrappers = function(getSegmentFromLine, getSegmentToLine, commentsById, allowCommenting) {
        var makeLineStart = function(segmentType) {

            var typeClass;
            var getSegmentLine;
            switch(segmentType) {
                case 'ADDED' :
                    typeClass = 'added-lines';
                    getSegmentLine = getSegmentToLine;
                    break;
                case 'REMOVED':
                    typeClass = 'deleted-lines';
                    getSegmentLine = getSegmentFromLine;
                    break;
                default:
                    typeClass = 'context';
                    getSegmentLine = getSegmentFromLine;
                    break;
            }

            // This should be extracted into a soy template in the future, but is not straightforward. mmcglynn has a spike for this.
            var startAndClass = '<div class="line ' + typeClass;
            var dataAttributesStart = '" data-line-type="' + segmentType + '" data-line="';
            var dataAttributesEnd = '">';
            var triggerStart = '<div class="add-comment-trigger"';
            var triggerEnd = '></div>';
            var end = '<div class="line-numbers"></div>' +
                '<code class="source">';
            return function (line, lineIndex) {
                return startAndClass + (line.truncated ? " truncated" : "") + (line.conflictMarker ? " conflict-" + line.conflictMarker.toLowerCase() : "") +
                    dataAttributesStart + (getSegmentLine() + lineIndex) + dataAttributesEnd +
                    triggerStart + (line.conflictMarker == 'marker' ? '' : ' title="' + stash_i18n('stash.web.diff.line.addcomment.tooltip', 'Add a comment to this line') + '"') + triggerEnd +
                    end;
            };
        };

        var makeLineEnd = function(segmentType) {
            var lineEnd =
                '</code>' +
            '</div>';
            var truncatedLineEnd =
                    '<span title="'+ stash_i18n('stash.web.diff.line.truncated', 'This line has been truncated.') + '">&hellip;</span>' +
                '</code>' +
            '</div>';

            var getSegmentLine;
            if (segmentType === 'ADDED') {
                getSegmentLine = getSegmentToLine;

            } else {
                getSegmentLine = getSegmentFromLine;
            }

            return function(line, lineIndex) {
                var comments;
                var normalLineEnd = line.truncated ? truncatedLineEnd : lineEnd;
                if (!line.commentIds ||
                    !line.commentIds.length ||
                    !(
                        comments = _.chain(line.commentIds).map(function(id) { return commentsById[id]; }).filter(_.identity).value()
                     ).length) {
                    return normalLineEnd;
                }
                return normalLineEnd + renderCommentContainer(getSegmentLine() + lineIndex, segmentType, comments, allowCommenting);
            };
        };

        return {
            ADDED : {
                lineStart : makeLineStart('ADDED'),
                lineEnd   : makeLineEnd('ADDED'),
                lineOffset : 1,
                emptyLine: ''
            },
            REMOVED : {
                lineStart : makeLineStart('REMOVED'),
                lineEnd   : makeLineEnd('REMOVED'),
                lineOffset: 1,
                emptyLine: ''
            },
            CONTEXT : {
                lineStart : makeLineStart('CONTEXT'),
                lineEnd   : makeLineEnd('CONTEXT'),
                lineOffset: 1,
                emptyLine: ''
            }
        };
    };

    function renderCommentContainer(lineNumber, lineType, comments, allowCommenting) {
        return comments.length === 0 ? '' : stash.feature.comments({
            line : lineNumber,
            lineType : lineType,
            comments : comments,
            currentUser : pageState.getCurrentUser(),
            extraClasses : 'comment-box',
            tips : tips
        });
    }

    function renderSegment(segment, lineWrappers) {
        var lineWrapper = lineWrappers[segment.type],
            lineStart = lineWrapper.lineStart,
            lineEnd = lineWrapper.lineEnd;
        return _.map(segment.lines, function(line, lineIndex) {
            return lineStart(line, lineIndex) +
                (line.line.length ? AJS.escapeHtml(line.line) : lineWrapper.emptyLine) +
                lineEnd(line, lineIndex);
        }).join('');
    }

    function getLineText(line) {
        return line.line;
    }

    function joinSegmentLines(segment) {
        // caches the joined text for use when rendering the ediff
        return (segment.fullText = _.map(segment.lines, getLineText).join('\n'));
    }

    function tokenizeSegment(segment) {
        // caches the tokens on the segment to pass to the ediff generator later
        return (segment.tokens = tokenizeString(joinSegmentLines(segment)));
    }

    function shouldRenderEdiff(segment, previousSegment) {
        // Determine if we should render an ediff for these segments based on the number of tokens they contain.
        // Number of tokens is cheap and gives us a rough idea of the *potential* to cause a slow ediff render (but does not guarantee it).
        // The cost of calculating a better metric (number of differences between segments) is roughly equivalent to rendering the ediffs anyway.
        // tokenizeSegment & joinSegmentLines cache the values generated to reuse when performing the actual render.

        return ((tokenizeSegment(segment).length < ediffTokenLimit) && (tokenizeSegment(previousSegment).length < ediffTokenLimit));
    }

    function renderHunkContent(hunk, commentsById, allowCommenting) {
        var fragments = [],
            segments = hunk.segments,
            l = segments.length,
            previousSegment = null,
            segment;

        var fromLine = hunk.sourceLine,
            toLine = hunk.destinationLine;

        var lineAwareWrappers = makeLineWrappers(
            function() { return fromLine; },
            function() { return toLine; },
            commentsById,
            allowCommenting
        );

        // Find adjacent REMOVED and ADDED segments. We only apply ediff to these segments
        for (var i = 0; i < l; i++) {
            segment = segments[i];
            if (!isEdiffDisabled && previousSegment && segment.type === 'ADDED' && shouldRenderEdiff(segment, previousSegment)) {
                // isEdiffDisabled is checked first because the tokenization operation is very expensive in IE8, need to bail out early.
                // shouldRenderEdiff is only called when the other conditions are true, so the same segment is never passed as both `segment` and `previousSegment`
                var ediffRegions = generateEdiffRegions(previousSegment.tokens, segment.tokens);

                fragments.push(
                    mergeStreams(previousSegment.fullText,
                        lineNodeStream(previousSegment.lines, getLineText, lineAwareWrappers[previousSegment.type]),
                        new DiffRegionNodeStream(ediffRegions.originalRegions)
                    )
                );
                fromLine += previousSegment.lines.length;
                previousSegment = null;

                fragments.push(
                    mergeStreams(segment.fullText,
                        lineNodeStream(segment.lines, getLineText, lineAwareWrappers[segment.type]),
                        new DiffRegionNodeStream(ediffRegions.revisedRegions)
                    )
                );
                toLine += segment.lines.length;
            } else {
                // All other segments are rendered using a simple join
                if (previousSegment) {
                    fragments.push(renderSegment(previousSegment, lineAwareWrappers));
                    fromLine += previousSegment.lines.length;
                    previousSegment = null;
                }
                // If we find a REMOVED segment, we defer it's rendering until we look at the next segment
                // to see if it's an add
                if (segment.type === 'REMOVED') {
                    previousSegment = segment;
                } else {
                    fragments.push(renderSegment(segment, lineAwareWrappers));
                    toLine += segment.lines.length;
                    if (segment.type === 'CONTEXT') { // context will increase both line counts.
                        fromLine += segment.lines.length;
                    }
                }
            }
        }
        // we need to ensure we dont forget to render the last segment i
        if (previousSegment) {
            fragments.push(renderSegment(previousSegment, lineAwareWrappers));
            fromLine += previousSegment.lines.length;
        }

        return '<div class="hunk" style="counter-reset: fromLines ' + (hunk.sourceLine-1) + ' toLines ' + (hunk.destinationLine-1) + ';">' +
                    fragments.join('') +
               '</div>';
    }

    var hunkAbove = '<div class="hunk-above"></div>',
        hunkBelow = '<div class="hunk-below"></div>';
    function getSeparatedHunkHtml(hunkHtmls) {
        return hunkBelow + hunkHtmls.join(hunkAbove + hunkBelow) + hunkAbove;
    }

    function renderDiffContent(diff, showComments, allowCommenting) {
        var commentsById = {};
        if (showComments) {
            _.each(diff.lineComments, function(comment) {
                commentsById[comment.id] = comment;
            });
        }

        return getSeparatedHunkHtml(_.map(diff.hunks, function(hunk) {
            return renderHunkContent(hunk, commentsById, allowCommenting);
        }));
    }

    function renderEmptyDiff(diff) {
        eve('stash.feature.diff-view.render.emptydiff', this);
        return stash.feature.fileContent.emptyDiff();
    }

    function renderErrors(data) {

        if (!data.errors || !data.errors.length) {
            return widget.aui.message.error({
                extraClasses : 'diff-view-message',
                contents : AJS.escapeHtml(stash_i18n('stash.web.repo.diff.unknown.error', 'An unknown error has occurred'))
            });
        }

        return _.map(data.errors, function(error) {
            return  widget.aui.message.error({
                extraClasses : 'diff-view-message',
                contents : AJS.escapeHtml(error.message)
            });
        }).join('');
    }

    function renderTooLargeDiff(commitRange, filePath) {
        var message;

        if(commitRange.getPullRequest()) {
            //TODO - Better message for pull request diffs that are too large to render. The user can't easily reproduce it on the command line.
            message = AJS.escapeHtml(stash_i18n('stash.web.pull-request.diff.toolarge.detail', 'The diff for this file is too large to render'));
        } else {
            var revisionId = commitRange.getUntilRevision() && commitRange.getUntilRevision().getId(),
                parentRevisionId = commitRange.getSinceRevision() && commitRange.getSinceRevision().getId();

            message = AJS.escapeHtml(
                stash_i18n('stash.web.repo.diff.toolarge.detail',
                    'The diff for this file is too large to render. You can still retrieve it manually with the following Git command.'));

            var gitCommand;
            if (parentRevisionId) {
                gitCommand = 'git diff -C -M ' + parentRevisionId + ' ' + revisionId + ' -- "' + filePath + '"';
            } else {
                gitCommand = 'git log -p ' + revisionId + ' -- "' + filePath + '"';
            }
            gitCommand = '<p class="scm-command">' + AJS.escapeHtml(gitCommand) + '</p>';

            message += gitCommand;
        }

        return widget.aui.message.warning({
            extraClasses : 'diff-view-message',
            contents : message
        });
    }

    function renderConflictMessage(conflict) {
        return stash.feature.fileContent.conflictMessage({
            titleHtml: AJS.escapeHtml(conflict.getConflictMessage()),
            messageHtml: AJS.escapeHtml(stash_i18n('stash.web.pull-request.diff.conflict.detail', 'This file is in a conflicted state. You will need to resolve the conflict manually before you can merge this pull request.'))
        });
    }

    DiffView._renderMethods = {
        hunk : {
            content : renderHunkContent
        },
        diff : {
            content : renderDiffContent
        },
        error : renderErrors,
        empty : renderEmptyDiff
    };

    return DiffView;
});
