define('layout/pull-request', [
    'jquery',
    'eve',
    'underscore',
    'memoir',
    'util/ajax',
    'util/navbuilder',
    'util/event',
    'model/page-state',
    'model/pull-request',
    'model/participant',
    'widget/approve',
    'widget/can-merge',
    'widget/merge-help',
    'widget/confirm-dialog',
    'feature/pull-request-edit',
    'widget/avatar-list',
    'exports'
], function(
    $,
    eve,
    _,
    memoir,
    ajax,
    nav,
    eventUtil,
    pageState,
    PullRequest,
    Participant,
    Approve,
    canMerge,
    mergeHelp,
    ConfirmDialog,
    PullRequestEdit,
    AvatarList,
    exports) {

    var pullRequest,
        $actionToolbar,
        $actionToolbarSpinner,
        $tabMenu;

    var haveKeyboardShortcutsObject = $.Deferred();

    function getActionUrl(action, withVersion) {
        var builder = nav.rest().currentPullRequest()[action]();

        if (withVersion) {
            builder = builder.withParams({ version: pageState.getPullRequest().getVersion() });
        }
        return builder.build();
    }

    function initActionButton($button, action) {
        $button.click({action: action}, buttonAction);
    }

    function initMergeButton($mergeButton) {
        if ($mergeButton.length) {
            var panelContent = stash.layout.pullRequest.pullRequestMergeDialogContent({
                sourceBranch: pullRequest.getFromRef().getDisplayId(),
                targetBranch: pullRequest.getToRef().getDisplayId(),
                person: pageState.getCurrentUser(),
                pullRequestId: pullRequest.getId()
            });

            var options = {
                buttonSelector : '.merge-pull-request',
                onShow : function(context, dialog) {
                    _.defer(function() { dialog.updateHeight(); });  // Need to allow for the content to render before resizing
                },
                confirmDialog : {
                    titleText: stash_i18n('stash.web.pull-request.merge.dialog.title', 'Merge Pull Request'),
                    panelContent: panelContent,
                    submitText : stash_i18n('stash.web.button.merge', 'Merge')
                },
                ajax : {
                    statusCode : {
                        '401' :  function(xhr, textStatus, errorThrown, errors, dominantError) {

                            return $.extend({}, dominantError, {
                                title: stash_i18n("stash.web.pull-request.merge.error.401.title", "Could not merge pull request"),
                                message: stash_i18n("stash.web.pull-request.merge.error.401.message", "You no longer have permission to merge this pull request."),
                                fallbackUrl: false,
                                shouldReload: true
                            });
                        },
                        '409' :  function(xhr, textStatus, errorThrown, errors, dominantError) {

                            var error = errors.errors && errors.errors.length && errors.errors[0];
                            if (error && (error.conflicted || (error.vetoes && error.vetoes.length))) {
                                eve('stash.pull-request.cant.merge', null, pullRequest, error.conflicted, error.vetoes);
                                eve('stash.pull-request.show.cant.merge.help');
                                return false;
                            }
                        }
                    }
                }
            };

            var mergeDialog = createActionDialog('merge', options);
            checkCanMerge($mergeButton, mergeDialog);
        }
    }

    function initDeclineButton($declineButton) {
        if ($declineButton.length) {
            var panelContent = "<p class='decline-message'>" +
                stash_i18n("stash.web.pull-request.decline.dialog.message",
                    "Are you sure you want to decline this pull request? This will close the pull request without merging.") + "</p>";

            var options = {
                buttonSelector : '.decline-pull-request',
                confirmDialog : {
                    titleText: stash_i18n('stash.web.pull-request.decline.dialog.title', 'Decline Pull Request'),
                    panelContent: panelContent,
                    submitText : stash_i18n('stash.web.button.decline', 'Decline')
                },
                ajax : {
                    statusCode : {
                        '401' :  function(xhr, textStatus, errorThrown, errors, dominantError) {
                            return $.extend({}, dominantError, {
                                title: stash_i18n("stash.web.pull-request.decline.error.401.title", "Could not decline pull request"),
                                message: stash_i18n("stash.web.pull-request.decline.error.401.message", "You no longer have permission to decline this pull request."),
                                fallbackUrl: false,
                                shouldReload: true
                            });
                        }
                    }
                }
            };

            createActionDialog('decline', options);
        }
    }

    function buttonAction(e) {
        e.preventDefault();

        if ($actionToolbar.hasClass('disabled')) {
            return;
        }
        $actionToolbar.addClass('disabled');
        $actionToolbarSpinner.show().spin('small');

        ajax.rest({
            url: getActionUrl(e.data.action, true),
            type: 'POST'
        }).done(function() {
            window.location.reload();
        }).fail(function() {
            $actionToolbar.removeClass('disabled');
            $actionToolbarSpinner.spinStop().hide();
        });
    }

    function createActionDialog(action, options) {
        var actionDialog = new ConfirmDialog($.extend({
            id : "pull-request-" + action + "-dialog",
            submitToHref : false,
            resizeOnShow : true,
            width : 650,
            height : 200
        }, options.confirmDialog));

        var currentXhr;

        actionDialog.addConfirmListener(function(promise, $trigger, removeDialog, dialog, $spinner) {
            actionDialog.setButtonsDisabled(true);
            $spinner.show();

            currentXhr = ajax.rest($.extend({
                url: getActionUrl(action, true),
                type: 'POST'
            }, options.ajax)).done(function() {
                window.location = nav.currentPullRequest().build();
            }).fail(function() {
                removeDialog();
            }).always(function() {
                currentXhr = null;
            });
        });

        actionDialog.addCancelListener(function(dialog) {
            if (currentXhr) {
                currentXhr.abort();
            }
        });

        actionDialog.attachTo(options.buttonSelector, options.onShow);

        return actionDialog;
    }

    function bindKeyboardShortcuts() {

        eve.on('stash.widget.keyboard-shortcuts.register-contexts', function(keyboardShortcuts) {
            keyboardShortcuts.enableContext('pull-request');
            haveKeyboardShortcutsObject.resolve(keyboardShortcuts);
        });

        eve.on('stash.keyboard.shortcuts.requestGotoPullRequestsListHandler', function(keys) {
            (this.execute ? this : AJS.whenIType(keys)).execute(function() {
                window.location.href = nav.currentRepo().allPullRequests().build();
            });
        });

        eve.on('stash.keyboard.shortcuts.requestChangePullRequestSectionHandler', function(keys) {
            (this.execute ? this : AJS.whenIType(keys)).execute(function(e) {
                var number = parseInt(String.fromCharCode(e.which), 10);
                var $link = $tabMenu.children().eq(number - 1).children('a');
                if (memoir.nativeSupport()) {
                    // trigger the pushstate events
                    $link.click();
                } else {
                    window.location.href = $link.prop('href');
                }
            });
        });
    }

    function checkCanMerge($mergeButton, mergeDialog) {
        var alreadyShown = false;
        eve.on('stash.pull-request.cant.merge', function(pullRequest, conflicted, vetoes) {
            if (!alreadyShown) {
                $mergeButton.attr('aria-disabled', 'true')
                            .attr('disabled', 'disabled')
                            .attr('title', stash_i18n('stash.web.pull-request.merge.issue.tooltip', 'This pull request can\'t be merged automatically.'));

                var title;
                if (conflicted && !vetoes.length) {
                    //Conflicted but nothing else wrong
                    title = stash.widget.pullRequest.mergeConflictTooltip();
                } else if ((conflicted && vetoes.length) || vetoes.length > 1) {
                    //Multiple vetoes or a conflict and at least one veto
                    title = stash.widget.pullRequest.mergeIssuesTooltip();
                } else {
                    //A single veto
                    title = vetoes[0].detailedMessage;
                }

                $('<div class="merge-conflict aui-icon-warning"/>').insertBefore($mergeButton.closest('.aui-toolbar2-secondary')).tooltip({
                    delayOut: 200,
                    gravity: 'n',
                    html: true,
                    hoverable: true,
                    offset: 7,
                    title: function() { return title; }
                });


                $(document).on('click', '.show-merge-help', function(e) {
                    e.preventDefault();
                    $('.tipsy').hide();
                    eve('stash.pull-request.show.cant.merge.help');
                });
                alreadyShown = true;
            }
        });

        canMerge(pullRequest);
    }

    function initTabs() {
        function setTabActive($tab) {
            $tab.addClass('active-tab')
                .siblings().removeClass('active-tab');
        }

        haveKeyboardShortcutsObject.done(function(keyboardShortcuts) {
            _.each($tabMenu.children(), function (tab, index) {
                var $tab = $(tab);
                var key = String(index + 1);
                var message = stash_i18n('stash.keyboard-shortcuts.pull-request.switch.tabs', 'Switch to the {0} tab', $tab.text());
                keyboardShortcuts.addCustomShortcut('pull-request', [ [ key ] ], message);
                $tab.attr('title', ($tab.attr('title') || message) + stash_i18n('stash.keyboard-shortcuts.type.x', " (Type ''{0}'')", key));
            });
        });

        if (memoir.nativeSupport()) {
            $tabMenu.on('click', 'a', function(e) {
                if (!eventUtil.openInSameTab(e)) {
                    return;
                }
                var $a = $(this),
                    $tab = $a.parent();

                if (!$tab.is('.active-tab')) {
                    setTabActive($tab);
                    eve('stash.layout.pull-request.urlRequested', null, $a.prop('href'));
                }

                e.preventDefault();
            });
            eve.on('stash.page.pull-request.view.contextLoaded', function(context) {
                setTabActive($tabMenu.find('[data-module-key="' + context.name + '"]'));
            });
        }
    }

    function initReviewersLists($reviewersList) {
        return new AvatarList($reviewersList, pullRequest.getReviewers(), {
            pullRequestId: pullRequest.getId()
        });
    }

    function initUpdateApproval() {

        function updateApproval(approvalJSON) {
            var reviewer = pullRequest.getReviewers().findByUser(approvalJSON.user);

            if (reviewer) {
                reviewer.setApproved(approvalJSON.approved);
                pullRequest.getReviewers().sort();  // Changing attributes does not trigger an automatic sort
            } else {
                var participant = pullRequest.getParticipants().findByUser(approvalJSON.user);

                if (participant) {
                    participant.setApproved(approvalJSON.approved);
                    pullRequest.getParticipants().sort();
                }
            }
        }

        eve.on('stash.widget.approve-button.adding', updateApproval);
        eve.on('stash.widget.approve-button.removing', updateApproval);
        eve.on('stash.widget.approve-button.add.failed', updateApproval);
        eve.on('stash.widget.approve-button.remove.failed', updateApproval);
    }

    function initParticipants() {

        function addParticipant(participantJSON) {
            var reviewer = pullRequest.getReviewers().findByUser(participantJSON.user);

            if (!reviewer) {
                var participant = pullRequest.getParticipants().findByUser(participantJSON.user);

                if (!participant) {
                    var newParticipant = new Participant(participantJSON);
                    pullRequest.addParticipant(newParticipant);

                    return true;
                }
            }

            return false;
        }

        eve.on('stash.widget.approve-button.adding', function(approvalJSON) {
            addParticipant({
                approved: approvalJSON.approved,
                user: approvalJSON.user,
                role: 'PARTICIPANT'
            });
        });

        eve.on('stash.feature.comments.comment.added', function(comment) {
            if (comment.author.name !== pullRequest.getAuthor().getUser().getName()) {
                var isNew = addParticipant({
                    approved: false,  // if the user has already approved, this won't remove it.
                    user: comment.author,
                    role: 'PARTICIPANT'
                });

                if (isNew && pageState.getCurrentUser() &&
                    comment.author.name === pageState.getCurrentUser().name) {
                    eve('stash.widget.watch-button.adding', this);
                }
            }
        });
    }

    exports.onReady = function(pullRequestJSON, isWatching) {
        pageState.setPullRequest(new PullRequest(pullRequestJSON));
        pageState.extend('isWatching');
        pageState.setIsWatching(isWatching);

        mergeHelp.init();

        pullRequest = pageState.getPullRequest();

        var $mergeButton = $('.merge-pull-request'),
            $declineButton = $('.decline-pull-request'),
            $reopenButton = $('.reopen-pull-request'),
            $approveButton = $('.approve'),
            pullRequestEdit = new PullRequestEdit(pullRequest);

        // Won't know which buttons exist on the page so add all of them
        $actionToolbar = $('.pull-request-toolbar .aui-toolbar2-secondary');
        $actionToolbarSpinner = $("<div class='spinner'/>").insertBefore($actionToolbar);

        $tabMenu = $('.content-body .aui-page-panel-content > .aui-tabs > .tabs-menu');

        initMergeButton($mergeButton);
        initDeclineButton($declineButton);
        initActionButton($reopenButton, 'reopen');
        pullRequestEdit.bind('.edit-pull-request, .add-description'); //.add-description is from pull-request-overview

        new Approve($approveButton, getActionUrl('approve'));

        initReviewersLists($('.reviewers'));
        initParticipants();
        initUpdateApproval();

        bindKeyboardShortcuts();

        initTabs();
    };
});
