define('page/admin/groupEdit',
    ['exports', 'jquery', 'util/ajax', 'util/navbuilder', 'widget/delete-dialog'],
    function (exports, $, ajax, navBuilder, DeleteDialog) {

        exports.onReady = function (readOnly, filterPickerSelector, addMoreSelector, deleteLink) {

            // dialog to confirm the deletion of the group
            DeleteDialog.bind(deleteLink,
                stash_i18n('stash.web.groups.delete', 'Delete group'),
                stash_i18n('stash.web.groups.delete.success', 'The group {0} was successfully deleted.'),
                stash_i18n('stash.web.groups.delete.fail', 'The group could not be deleted.'),
                function(name) {
                    window.location = navBuilder.admin().groups().deleteSuccess(name).build();
                    return false; // don't notify on view page, wait for page-pop
                }, function() {
                    return $('#group-name').text();
                });

            // ---------------------------------------------------------------------
            // Group Edit Pickers
            // (taken without changes from the Embedded Crowd admin user plugin,
            // except for the conversion to Stash REST + error reporting)
            // ---------------------------------------------------------------------

            var pickers = [];

            /**
             * @param context the context entity (a username or group name)
             * @param moreItemsUrl the URL of the resource that provides more items (users or groups) for the filtered picker
             * @param actionUrl the URL of the resource that adds or removes items from the context entity
             * @param renderName true if the entity's name should be rendered as part of the item content (in parentheses)
             * @param xsrfParam the name of the xsrf-protection form parameter
             * @param xsrfToken the xsrf-protection token value
             *
             * @return an initialised options object for a jquery.filteredpicker control
             */
            function baseFilterPickerOptions(context, moreItemsUrl, actionUrl, renderName, xsrfParam, xsrfToken, linkToItemDetails) {
                return {
                    pageSize: 25,
                    infiniteScroll: false,
                    buttonLoadMoreText: stash_i18n('stash.picker.load.more', "Load more"),
                    retrieveItems: function(filter, offset, pageSize, callback) {
                        ajax.rest({
                            url: moreItemsUrl,
                            type: 'GET',
                            data: {
                                context: context,
                                filter: filter,
                                start: offset,
                                pageSize: pageSize,
                                avatarSize: stash.widget.avatarSizeInPx({ size: 'xsmall' })
                            }
                        }).done(function (data) {
                            var users = [];
                            for (var i = 0; i < data.values.length; i++) {
                                var item = data.values[i];
                                var content = $("<div>");
                                var avatar = stash.widget.avatar({
                                    person: item,
                                    size: 'xsmall'
                                });
                                content
                                    .append(avatar)
                                    .append(
                                        $(linkToItemDetails ?
                                            '<a href="' + navBuilder.admin().users().view(item.name).build() + '"></a>' :
                                            '<span></span>'
                                        )
                                            .addClass('display-name')
                                            .text(item.displayName)
                                    );
                                if (renderName) {
                                    content.append($("<div class='name'>").text(item.name));
                                }
                                users.push({
                                    content: content,
                                    id: item.name,
                                    disabled: !item.mutableGroups,
                                    disabledReason: item.mutableGroups ? null : stash_i18n("stash.web.user.error.can.not.move.read.only.user", "User is from a read-only directory.")
                                });
                            }
                            callback(users, !data.isLastPage);
                        });
                    },
                    buttonPressed: function(id, doneCallback, errorCallback) {
                        ajax.rest({
                            url: actionUrl,
                            type: 'POST',
                            data: {
                                context: context,
                                itemName: id
                            },
                            statusCode : {
                                '400' : function() { return false; },
                                '403' : function() { return false; },
                                '404' : function() { return false; },
                                '409' : function() { return false; },
                                '500' : function() { return false; }
                            }
                        }).done(doneCallback
                         ).fail(function (xhr, textStatus, errorThrown, data) {
                            if (data && data.errors) {
                                errorCallback(data.errors);
                            } else {
                                errorCallback([{'message': stash_i18n("stash.web.groups.picker.unexpected.error", "An unexpected error occurred.")}]);
                            }
                        });
                    },
                    noMoreItemsText: stash_i18n("stash.web.users.picker.none.added", "This group has no users."),
                    noItemsText: stash_i18n("stash.web.users.picker.none.matching", "No matching users."),
                    loadingItemsText: stash_i18n("stash.web.users.picker.loading", "Loading users...")
                };
            }

            /**
             * @return an enhanced set of jquery.filteredpicker options suitable for rendering a modal filteredpicker
             * control
             *
             * @see baseFilterPickerOptions
             */
            function baseModalFilterPickerOptions(context, moreItemsUrl, actionUrl, renderName, xsrfHeader, xsrfToken) {
                var options = baseFilterPickerOptions(context, moreItemsUrl, actionUrl, renderName, xsrfHeader, xsrfToken, false);
                $.extend(options, {
                    infiniteScroll: true,
                    panelClass: 'user-dialog-panel',
                    onClose: function() {
                        for (var i = 0; i < pickers.length; i++) {
                            pickers[i].refreshPicker();
                        }
                    },
                    noMoreItemsText: stash_i18n("stash.web.users.picker.all.added", "This group already contains all users.")
              });
                return options;
            }
            /**
             * Add i18n properties to a jquery.filteredpicker options object suitable for rendering when the control is
             * adding or removing users to a group.
             *
             * @param isAdding true if this picker is for adding users to a group, or false if this picker is for
             *        removing users from a group
             * @param options the options object to enhance with i18n properties
             *
             * @return the supplied options object
             */
            function withGroupContextOptions(isAdding, options) {
                $.extend(options, {
                    title: isAdding ? stash_i18n("stash.web.groups.picker.users.available", "Available Users") : stash_i18n("stash.web.groups.picker.users.assigned", "Assigned Users"),
                    filterPlaceholder: stash_i18n("stash.web.groups.picker.users.placeholder", "Search users"),
                    showButton: readOnly,
                    buttonText: isAdding ? stash_i18n("stash.web.groups.picker.users.add", "Add user") : stash_i18n("stash.web.groups.picker.users.remove", "Remove User"),
                    buttonDoingText: isAdding ? stash_i18n("stash.web.groups.picker.adding", "Adding...") : stash_i18n("stash.web.groups.picker.removing", "Removing..."),
                    buttonDoneText: isAdding ? stash_i18n("stash.web.groups.picker.added", "Added") : stash_i18n("stash.web.groups.picker.removed", "Removed"),
                    buttonDoneClass: isAdding ? "green" : "red",
                    closeButtonText: stash_i18n("stash.web.groups.picker.done", "Done")
                });
                return options;
            }

            /**
             * Bind the picker on the group edit page for viewing (and removing) members of a group.
             */
            $(filterPickerSelector).each(function() {
                var $pickerContainer = $(this);
                pickers.push(
                    $pickerContainer.filteredPicker(
                        withGroupContextOptions(
                            false,
                            baseFilterPickerOptions(
                                $pickerContainer.data("group"),
                                $pickerContainer.data("more-url"),
                                $pickerContainer.data("remove-url"),
                                true,
                                $pickerContainer.data("xsrf-token-name"),
                                $pickerContainer.data("xsrf-token-value"),
                                true
                            )
                        )
                    )
                );
            });

            /**
             * Bind the modal picker on the group edit page for adding members to a group.
             */
            $(addMoreSelector).each(function() {
                var $pickerTrigger = $(this);
                $pickerTrigger.bindModalFilteredPicker(
                    withGroupContextOptions(
                        true,
                        baseModalFilterPickerOptions(
                            $pickerTrigger.data("group"),
                            $pickerTrigger.data("more-url"),
                            $pickerTrigger.data("add-url"),
                            true,
                            $pickerTrigger.data("xsrf-token-name"),
                            $pickerTrigger.data("xsrf-token-value")
                        )
                    )
                );
            });

        };
});
