define('page/admin/userEdit', [
    'jquery',
    'underscore',
    'util/ajax',
    'util/error',
    'util/navbuilder',
    'widget/delete-dialog',
    'widget/confirm-dialog',
    'widget/submit-spinner',
    'exports'
], function (
    $,
    _,
    ajax,
    errorUtil,
    navBuilder,
    DeleteDialog,
    ConfirmDialog,
    SubmitSpinner,
    exports
) {

        function notify(content) {
            var $notification = $('.content-body .notifications');
            $notification.empty().html(content);
        }

        function notifySuccess(message) {
            notify(widget.aui.message.success({ contents: message }));
        }

        function setErrorSpan(fieldSelector, message) {
            $(fieldSelector).parent(".field-group").append($("<span class='error'></span>").text(message));
        }

        function clearErrors() {
            $('.panel-details .error, .content-body .notifications > .error').remove();
        }

        function notifyErrors(errors) {
            if (_.isArray(errors)) {
                _.each(errors, function(error) {
                    if (error.message && error.context && error.context === 'email') {
                        setErrorSpan("#email", error.message);
                    } else if (error.message && error.context && error.context === 'displayName') {
                        setErrorSpan("#fullname", error.message);
                    } else if (error.message) {
                        notify(widget.aui.message.error({ contents: error.message }));
                    } else {
                        notify(widget.aui.message.error({ contents: error }));
                    }
                });
            } else if (_.isString(errors)) {
                notify(widget.aui.message.error({ contents: errors }));
            }
        }

        // dialog to confirm the deletion of the user
        function initialiseDeleteDialog(deleteLink) {
            DeleteDialog.bind(deleteLink,
                stash_i18n('stash.web.users.delete', 'Delete user'),
                stash_i18n('stash.web.users.delete.success', 'The user {0} was successfully deleted.'),
                stash_i18n('stash.web.users.delete.fail', 'The user could not be deleted.'),
                function(name) {
                    window.location = navBuilder.admin().users().deleteSuccess(name).build();
                    return false; // don't notify on view page, wait for page-pop
                }, function() {
                    return $('#fullname').val();
                });
        }

        // dialog to change the password
        function initialisePasswordDialog(passwordLink) {
            $(passwordLink).click(function(e) {
                var addError = function ($field, message) {
                    $field.after($('<div></div>').text(message).addClass('error'));
                };

                var $dialogContent = $('#password-dialog-template').clone().removeAttr('id');
                $('[id]', $dialogContent).each(function() {
                    $(this).prev('label').attr('for', this.id = this.id + '-dialog');
                });

                var dialog = new AJS.Dialog({ width:433, id:'change-password-dialog', closeOnOutsideClick: false,
                    keypressListener: function (e) {
                        e.stopImmediatePropagation(); // AUIDialog.updateHeight() rebinds the keypressListener at every call, even if it's already bound to the event;
                                                      // thus we need to have jQuery stops the immediate propagation of the event to prevent successive invocations.
                                                      // For example, the sequence dialog.show().updateHeight().updateHeight() would have the handler bound three times.
                        if (e.keyCode === $.ui.keyCode.ENTER) {
                            $('.button-panel-submit-button', this).click();
                        } else if (e.keyCode === $.ui.keyCode.ESCAPE) {
                            dialog.remove();
                        }
                    }
                });
                dialog.addHeader(stash_i18n('stash.web.users.change.password.dialog', 'Change the password for {0}', $('<p></p>').text(($('#name', '.panel-details form').val())).html()))
                    .addPanel('content', $dialogContent)
                    .addSubmit(stash_i18n('stash.web.button.save', 'Save'), function(dialog,page) {
                        var $form = $('form', $dialogContent);
                        var $spinner = new SubmitSpinner($('.button-panel-submit-button', dialog.getPage(0).buttonpanel), 'before').show();

                        var $oldPassword = $form.find('input[name=oldPassword]');
                        var $password = $form.find('input[name=password]');
                        var $passwordConfirm = $form.find('input[name=passwordConfirm]');

                        dialog.disable();  // Prevent double submission
                        var buttonPanel = dialog.getPage(0).buttonpanel;
                        buttonPanel.addClass('disabled');

                        ajax.rest({
                            url: $form.attr('action'),
                            type: 'PUT',
                            data: ajax.formToJSON($form),
                            statusCode : {
                                '*' : function() { return false; /* this is already a popup: handle all the errors locally */ }
                            }
                        }).always(function() {
                            $spinner.remove();
                        }).done(function () {
                            dialog.remove();
                            notifySuccess(stash_i18n('stash.web.users.password.update.success', 'Password successfully updated.'));
                        }).fail(function (xhr, textStatus, errorThrown, data) {
                            dialog.enable();
                            buttonPanel.removeClass('disabled');

                            $form.find('div.error').remove();
                            if (data && data.errors && data.errors[0] && data.errors[0].message) {
                                _.each(data.errors, function(error) {
                                    if (error.context && error.message) {
                                        // check each context manually, rather than using a selector based on it, which would permit to inject some text anywhere in the page
                                        if (error.context === "oldPassword") {
                                            addError($oldPassword, error.message);
                                        } else if (error.context === "password") {
                                            addError($password, error.message);
                                        } else if (error.context === "passwordConfirm") {
                                            addError($passwordConfirm, error.message);
                                        }
                                    } else if (error.message) {
                                        addError($passwordConfirm, error.message);
                                    }
                                });
                            } else {
                                addError($passwordConfirm, stash_i18n('stash.web.users.change.password.failure', 'The password could not be changed.'));
                            }
                            dialog.updateHeight();
                        });

                    })
                    .addCancel(stash_i18n('stash.web.button.cancel', 'Cancel'), function(dialog,page) {
                        dialog.remove();
                    })
                    .show()
                    .updateHeight();
                return false;
            });
        }

        // form for editing user details
        function initialiseForm() {

            // utility functions
            function rollback($form) {
                $form.find('input[type=text]').each(function() {
                    var $this = $(this);
                    $this.val($this.attr('rollback'));
                });
            }
            function updateDetails($form, data) {
                $form.find('#fullname').val(data.displayName);
                $form.find('#email').val(data.emailAddress);
                $form.find('input[type=text]').each(function() {
                    var $this = $(this);
                    $this.attr('rollback', $this.val());
                });
            }
            function closeEditDetails($form) {
                $form.removeClass('editing').find('#fullname, #email').attr('readonly', 'readonly');
                $('#ajax-status-message').empty();
                clearErrors();
            }

            // event bindings
            $('#edit-details').click(function(e) {
                $('.panel-details form.editable').addClass('editing').find('#fullname, #email').removeAttr('readonly');
                if (e.target.id !== 'email') {
                    $('#fullname', '.panel-details form.editable').focus();
                }
                e.preventDefault();
            });
            $('.panel-details form.editable').keyup(function(e) {
                if(e.which == $.ui.keyCode.ENTER) {
                    $('.save', this).click();
                } else if (e.which == $.ui.keyCode.ESCAPE) {
                    $('a.cancel', this).click();
                }
            });
            $('.cancel', '.panel-details form.editable').click(function(e) {
                e.preventDefault();
                var $form = $(this).parents('form');
                rollback($form);
                closeEditDetails($form);
                return false;
            });
            $('.save', '.panel-details form.editable').click(function(e) {
                e.preventDefault();
                clearErrors();
                var $form = $(this).parents('form');
                var displayName = $form.find('#fullname').val();
                ajax.rest({
                    url: $form.attr('action'),
                    type: 'PUT',
                    data: {
                        name: $form.find('#name').val(),
                        displayName: displayName,
                        email: $form.find('#email').val()
                    },
                    statusCode : { // these errors are handled locally.
                        '500' : function() {
                            return false;
                        },
                        '404' : function() {
                            return false;
                        },
                        '401' : function() {
                            return false;
                        },
                        '400' : function() {
                            return false;
                        }
                    }
                }).done(function(data) {
                    updateDetails($form, data);
                    closeEditDetails($form);
                    notifySuccess(stash_i18n('stash.web.users.update.success', '{0}{1}{2} was successfully updated.', "<strong>", AJS.escapeHtml(displayName), "</strong>"));
                }).fail(function (xhr, textStatus, errorThrown, data) {
                    var errors = (data && data.errors) ? data.errors : stash_i18n('stash.web.users.update.failure', 'The user could not be updated.');
                    notifyErrors(errors);
                });
            });
        }

        function initialiseUserPicker(filterPickerSelector, addMoreSelector, displayGroupLinks) {

            // ---------------------------------------------------------------------
            // User Edit Pickers
            // (taken without changes from the Embedded Crowd admin user plugin,
            // except for the conversion to Stash REST + error reporting)
            // ---------------------------------------------------------------------

            // TODO: move duplicate code with groupedit.js to a dedicated feature
            var pickers = [];

            /**
             * @param context the context entity (a username or group name)
             * @param moreItemsUrl the URL of the resource that provides more items (users or groups) for the filtered picker
             * @param actionUrl the URL of the resource that adds or removes items from the context entity
             * @param renderName true if the entity's name should be rendered as part of the item content (in parentheses)
             * @param xsrfParam the name of the xsrf-protection form parameter
             * @param xsrfToken the xsrf-protection token value
             *
             * @return an initialised options object for a jquery.filteredpicker control
             */
            function baseFilterPickerOptions(context, editable, moreItemsUrl, actionUrl, renderName, xsrfParam, xsrfToken, linkToItemDetails) {
                return {
                    pageSize: 25,
                    infiniteScroll: false,
                    buttonLoadMoreText: stash_i18n('stash.picker.load.more', "Load more"),
                    retrieveItems: function(filter, offset, pageSize, callback) {
                        ajax.rest({
                            url: moreItemsUrl,
                            type: 'GET',
                            data: {
                                context: context,
                                filter: filter,
                                start: offset,
                                pageSize: pageSize
                            }
                        }).done(function (data) {
                            var users = [];
                            for (var i = 0, len = data.values.length; i < len; i++) {
                                var item = data.values[i];
                                var content = $("<div>");
                                content.append(
                                    $(linkToItemDetails ?
                                        '<a href="' + navBuilder.admin().groups().view(item.name).build() + '"></a>' :
                                        '<span></span>'
                                    )
                                        .addClass('display-name')
                                        .text(item.name)
                                        .attr("title", item.name)
                                );
                                users.push({
                                    content: content,
                                    id: item.name,
                                    disabled: !editable,
                                    disabledReason: editable ? null : stash_i18n("stash.web.user.error.can.not.move.read.only.user", "User is from a read-only directory.")
                                });
                            }
                            callback(users, !data.isLastPage);
                        });
                    },
                    buttonPressed: function(id, doneCallback, errorCallback) {
                        ajax.rest({
                            url: actionUrl,
                            type: 'POST', // TODO: should be a delete when the user is removed from the group
                            data: {
                                context: context,
                                itemName: id
                            },
                            statusCode : {
                                '400' : function() { return false; },
                                '403' : function() { return false; },
                                '404' : function() { return false; },
                                '409' : function() { return false; },
                                '500' : function() { return false; }
                            }
                        }).done(doneCallback
                         ).fail(function (xhr, textStatus, errorThrown, data) {
                            if (data && data.errors) {
                                errorCallback(data.errors);
                            } else {
                                errorCallback([{'message': stash_i18n("stash.web.groups.picker.unexpected.error", "An unexpected error occurred.")}]);
                            }
                        });
                    },
                    noMoreItemsText: stash_i18n("stash.web.groups.picker.none.added", "This user is not a member of any group."),
                    noItemsText: stash_i18n("stash.web.groups.picker.none.matching", "No matching groups."),
                    loadingItemsText: stash_i18n("stash.web.groups.picker.loading", "Loading groups...")
                };
            }

            /**
             * @return an enhanced set of jquery.filteredpicker options suitable for rendering a modal filteredpicker
             * control
             *
             * @see baseFilterPickerOptions
             */
            function baseModalFilterPickerOptions(context, editable, moreItemsUrl, actionUrl, renderName, xsrfHeader, xsrfToken) {
                var options = baseFilterPickerOptions(context, editable, moreItemsUrl, actionUrl, renderName, xsrfHeader, xsrfToken, false);
                $.extend(options, {
                    infiniteScroll: true,
                    panelClass: 'group-dialog-panel',
                    onClose: function() {
                        for (var i = 0; i < pickers.length; i++) {
                            pickers[i].refreshPicker();
                        }
                    },
                    noMoreItemsText: stash_i18n("stash.web.groups.picker.all.added", "This user is already a member of all groups.")
                });
                return options;
            }

            /**
             * Add i18n properties to a jquery.filteredpicker options object suitable for rendering when the control is
             * adding or removing users to a group.
             *
             * @param isAdding true if this picker is for adding users to a group, or false if this picker is for
             *        removing users from a group
             * @param options the options object to enhance with i18n properties
             *
             * @return the supplied options object
             */
            function withGroupContextOptions(isAdding, options) {
                $.extend(options, {
                    title: isAdding ? stash_i18n("stash.web.groups.picker.users.available", "Available Users") : stash_i18n("stash.web.groups.picker.users.assigned", "Assigned Users"),
                    filterPlaceholder: stash_i18n("stash.web.groups.picker.users.placeholder", "Search users"),
                    buttonText: isAdding ? stash_i18n("stash.web.groups.picker.users.add", "Add user") : stash_i18n("stash.web.groups.picker.users.remove", "Remove user"),
                    buttonDoingText: isAdding ? stash_i18n("stash.web.groups.picker.adding", "Adding...") : stash_i18n("stash.web.groups.picker.removing", "Removing..."),
                    buttonDoneText: isAdding ? stash_i18n("stash.web.groups.picker.added", "Added") : stash_i18n("stash.web.groups.picker.removed", "Removed"),
                    buttonDoneClass: isAdding ? "green" : "red",
                    closeButtonText: stash_i18n("stash.web.groups.picker.done", "Done")
                });
                return options;
            }

            /**
             * Add i18n properties to a jquery.filteredpicker options object suitable for rendering when the control is
             * adding or removing groups to a user.
             *
             * @param isAdding true if this picker is for adding groups to a user, or false if this picker is for removing
             *        groups from a user
             * @param options the options object to enhance with i18n properties
             *
             * @return the supplied options object
             */
            function withUserContextOptions(isAdding, options) {
                $.extend(options, {
                    title: isAdding ? stash_i18n("stash.web.users.picker.groups.available", "Available Groups") : stash_i18n("stash.web.users.picker.groups.assigned", "Assigned Groups"),
                    filterPlaceholder: stash_i18n("stash.web.users.picker.groups.placeholder", "Search groups"),
                    buttonText: isAdding ? stash_i18n("stash.web.users.picker.groups.add", "Add Group") : stash_i18n("stash.web.users.picker.groups.remove", "Remove Group"),
                    buttonDoingText: isAdding ? stash_i18n("stash.web.users.picker.adding", "Adding...") : stash_i18n("stash.web.users.picker.removing", "Removing..."),
                    buttonDoneText: isAdding ? stash_i18n("stash.web.users.picker.added", "Added") : stash_i18n("stash.web.users.picker.removed", "Removed"),
                    buttonDoneClass: isAdding ? "green" : "red",
                    closeButtonText: stash_i18n("stash.web.users.picker.done", "Done")
                });
                return options;
            }

            /**
             * Bind the picker on the user edit page for viewing (and removing) groups from a user.
             */
            $(filterPickerSelector).each(function() {
                var $pickerContainer = $(this);
                pickers.push(
                    $pickerContainer.filteredPicker(
                        withUserContextOptions(
                            false,
                            baseFilterPickerOptions(
                                $pickerContainer.data("user"),
                                $pickerContainer.data("editable"),
                                $pickerContainer.data("more-url"),
                                $pickerContainer.data("remove-url"),
                                false,
                                $pickerContainer.data("xsrf-token-name"),
                                $pickerContainer.data("xsrf-token-value"),
                                displayGroupLinks
                            )
                        )
                    )
                );
            });

            /**
             * Bind the modal picker on the user edit page for adding the user to groups.
             */
            $(addMoreSelector).each(function() {
                var $pickerTrigger = $(this);
                $pickerTrigger.bindModalFilteredPicker(
                    withUserContextOptions(
                        true,
                        baseModalFilterPickerOptions(
                            $pickerTrigger.data("user"),
                            true,
                            $pickerTrigger.data("more-url"),
                            $pickerTrigger.data("add-url"),
                            false,
                            $pickerTrigger.data("xsrf-token-name"),
                            $pickerTrigger.data("xsrf-token-value")
                        )
                    )
                );
            });
        }


        exports.onReady = function (filterPickerSelector, addMoreSelector, passwordLink, deleteLink, displayGroupLinks) {

            initialiseDeleteDialog(deleteLink);
            initialisePasswordDialog(passwordLink);
            initialiseForm();
            initialiseUserPicker(filterPickerSelector, addMoreSelector, displayGroupLinks);

            $(document).ready(function() {
                if (location.hash) {
                    $('.menu-item > a[href="' + location.hash + '"]').click();
                }
            });

        };
    }
);
