define('util/client-storage', [
    'jquery',
    'underscore',
    'model/page-state',
    'exports'
], function(
    $,
    _,
    pageState,
    exports
    ) {

    var SESSION = 'session',
        LOCAL = 'local';

    var componentDelimiter = "_",
        cleanupKey = 'lastCleanup',
        hasCheckedCleanUpKey = 'hasCheckedCleanUp',
        oneMonth = 1000 * 60 * 60 * 24 * 30,
        hasCheckedCleanUp = !!(getRawItem(hasCheckedCleanUpKey, SESSION));

    function buildKey(components, context) {
        if (_.isString(components)) {
            components = [components];
        }

        if (!_.isArray(components)) {
            throw Error('keyBuilder requires an array of components');
        }

        if (context) {
            // Add the context to the key
            components.push(context);

            // This switch falls through each level adding progressively more detail the higher up you start
            // e.g. the user context just adds the current user's username,
            // but `pull-request` adds the pull request id, the repo slug, the project key and the current username
            // This may need refactoring if we introduce new contexts that don't fit into this waterfall.
            switch(context) {
                case 'pull-request':
                    components.push(pageState.getPullRequest() && pageState.getPullRequest().getId());
                    /* falls through */
                case 'repo':
                    components.push(pageState.getRepository() && pageState.getRepository().slug);
                    /* falls through */
                case 'project':
                    components.push(pageState.getProject() && pageState.getProject().key);
                    /* falls through */
                case 'user':
                    components.push(pageState.getCurrentUser() && pageState.getCurrentUser().name);
                break;
            }
        }

        return components.join(componentDelimiter);
    }

    function getRawItem(key, type) {
        //Get the entire JSON object from localStorage.
        //Use if you want to access the metadata of an entry
        var rawItem,
            item;

        rawItem = window[(type || LOCAL) + "Storage"].getItem(key);

        try {
            item = JSON.parse(rawItem);
        } catch(exception) {
            //rawItem was invalid JSON, so just return the raw string
            item = rawItem;
        }

        return item;
    }

    function getItem(key, type) {
        //Return the `data` attribute of the JSON object stored in localStorage, or the raw value if it's not wrapped in a object.
        //`type` is LOCAL by default
        var item = getRawItem(key, type);
        return $.isPlainObject(item) && _.has(item, 'data') ? item.data : item;
    }

    function getSessionItem(key) {
        return getItem(key, SESSION);
    }

    function setRawItem(key, obj, type) {
        //Save the object as is to client storage, don't add meta data
        window[(type || LOCAL) + "Storage"].setItem(key, JSON.stringify(obj));
    }

    function setItem(key, obj, extraProperties, type) {
        //Don't allow extraProperties to overwrite the core attributes, `data` and `timestamp`;
        //Currently the only useful extraProperty is Boolean `noCleanup`
        //`type` is LOCAL by default
        var item = _.extend({}, extraProperties, {
            timestamp: new Date().getTime(),
            data: obj
        });

        setRawItem(key, item, type);

        if (!type || type === LOCAL) {
            //Defer cleanup task until after the calling code has finished executing
            _.defer(checkCleanup);
        }
    }

    function setSessionItem(key, obj, extraProperties) {
        setItem.call(this, key, obj, extraProperties, SESSION);
    }

    function removeItem(key, type){
        window[(type || LOCAL) + "Storage"].removeItem(key);
    }

    function removeSessionItem(key) {
        removeItem(key, SESSION);
    }

    function checkCleanup(){
        if (hasCheckedCleanUp) {
            //Short circuit if we have already checked for cleanup this page/session
            return;
        }

        var lastCleanup = getRawItem(cleanupKey);

        if (!lastCleanup || new Date().getTime() - lastCleanup > oneMonth) {
            doCleanup();
        }

        hasCheckedCleanUp = true; //Prevent reruns of the cleanup check in this instantiation of client-storage
        setRawItem(hasCheckedCleanUpKey, true, SESSION); //Prevent reruns of the cleanup check for the life of this session
    }

    function doCleanup(){
        var currTime = new Date().getTime();
        _.each(_.keys(localStorage), function(key){
            if (key !== cleanupKey) {
                //don't cleanup the cleanup tracker
                var item = getRawItem(key);
                if (item.timestamp && !item.noCleanup && (currTime - item.timestamp > oneMonth)) {
                    removeItem(key);
                }
            }
        });
        setRawItem(cleanupKey, new Date().getTime());
    }

    exports.LOCAL = LOCAL;
    exports.SESSION = SESSION;
    exports.buildKey = buildKey;
    exports.getItem = getItem;
    exports.getSessionItem = getSessionItem;
    exports.setItem = setItem;
    exports.setSessionItem = setSessionItem;
    exports.removeItem = removeItem;
    exports.removeSessionItem = removeSessionItem;
});