/**
 * This module holds functions for improving the performance of JS code.
 * Currently it contains queueDOMRead and queueDOMWrite, which can be used to ensure all DOM reads are executed at
 * once, and all DOM writes are executed at once to avoid reflows that occur when reads and writes are interleaved.
 *
 * See also: http://www.stubbornella.org/content/2009/03/27/reflows-repaints-css-performance-making-your-javascript-slow/
 */
define('util/performance', ['jquery'], function ($) {


    /**
     * This creates a queue of function callbacks. When the first callback is added (using the queue function),
     * the whole queue is registered to be invoked as soon as the current JS event loop finishes executing
     * (with a setTimeout(...,0)). If any other callbacks get registered in the meantime, they'll all be executed right
     * after the first one. Then the queue will reset itself to empty and await the next callback to be added. The
     * queue takes in a callback (executedCallback) that will be called whenever the queue is executed and reset.
     *
     * You can also force immediate queue execution without waiting for the setTimeout (forceExecute). Below I use
     * this to avoid superfluous setTimeouts. If the Read queue executes, it will call the Write queue immediately
     * afterwards, which will call the Read queue in case anything was added during the Write callbacks, and so forth
     * until there are no more new callbacks being added to either queue.
     * @param {Function} executedCallback a callback which will be called whenever the queue is executed.
     * @return {Object}
     */
    function getDelayedExecutionQueue(executedCallback) {
        var queue = new $.Callbacks();
        var executing = false;
        var pendingExecution = null;

        function execute() {
            executing = true;
            // execute all the pending callbacks
            queue.fire();
            // reset the queue
            queue = new $.Callbacks();
            executing = false;

            pendingExecution = null;

            // call the callback since we just executed the queue.
            executedCallback();
        }

        return {
            /**
             * If there are callbacks scheduled to be executed, remove the scheduled execution, execute them now
             * and reset the queue.
             */
            forceExecute : function() {
                if (pendingExecution) {
                    clearTimeout(pendingExecution);
                    execute();
                }
            },
            /**
             * Add a callback to the queue. If there is no scheduled execution for the queue, schedule one.
             * @param fn {Function} the callback to add
             */
            queue : function(fn) {
                if (executing) { // if we're currently executing, there is no harm executing this immediately.
                    fn();
                } else { // otherwise, add it to the queue for executing later.
                    queue.add(fn);

                    if (!pendingExecution) { // if there isn't an execution scheduled yet, schedule one.
                        pendingExecution = setTimeout(execute, 0);
                    }
                }
            }
        };
    }

    // create a separate queue for DOM reads and DOM writes. Avoid the overhead of setTimeout by forcing immediate
    // execution of the partner queue when all your own callbacks have finished.
    var DOMReadQueue = getDelayedExecutionQueue(function() { DOMWriteQueue.forceExecute(); });
    var DOMWriteQueue = getDelayedExecutionQueue(function() { DOMReadQueue.forceExecute(); });

    return {
        queueDOMRead : DOMReadQueue.queue,
        queueDOMWrite : DOMWriteQueue.queue
    };
});
