/**
 * Provides widgets for defining filtered pickers.
 *
 * Dependencies: AUI (the modal filtered picker displays in an AUI dialog)
 *
 * Usage:
 * <pre>
 *
 * </pre>
 *
 * modalFilteredPicker()
 *      trigger - this element will spawn the picker in a modal dialog when clicked
 *      options - see below
 *
 * filteredPicker() makes a filtered picker into a container.
 *      options - see below
 *
 *      options:
 *          - filterPlaceholder String
 *          - panelClass String
 *          - showButton boolean whether to show the button on each entry
 *          - buttonText String
 *          - buttonPressedText String
 *          - buttonBusyClass class to apply while the button action is occurring
 *          - buttonDoneClass class to apply when the button action has completed
 *          - title String (only used for the modal filtered picker, this is the title of the modal dialog)
 *          - doneButtonText String (only used for the modal filtered picker, this button is used to close the picker)
 *          - pageSize the number of items to initially retrieve
 *          - retrieveItems(filter, offset, pageSize, callback)
 *                  function called when the filtered text changes,
 *                  the callback takes two arguments:
 *                      an array of { id: String, content: Object }, which the filter results. Note that the object
 *                          object MUST BE PROPERLY ESCAPED as it is allowed to contain HTML and will be appended
 *                          directly to the DOM
 *                      a boolean, true if there are more items that match the filter, that are not shown
 *          - noItemsText String displayed if there are no items or no items matching the current filter
 *          - loadingItemsText String displayed when items are being loaded from the server
 *          - buttonPressed(id, doneCallback, errorCallback) function called when a button is pressed with the id of the row
 *                    doneCallback (no-args) should be called when the operation has completed successfully
 *                    errorCallback should be called with a String error message if the operation failed
 *          - onClose() (optional) function (only used for the modal filtered picker) called when the dialog is closed
 *          - isDisabled() (optional) function that returns true if the picker's trigger is currently disabled, and the picker
 *            should not be launched
 */
(function($, _) {

    var cleanOptions = function(options) {
        if (!options.pageSize) {
            options.pageSize = 20;
        }
        if (options.showFilter === undefined) {
            options.showFilter = true;
        }
        if (options.infiniteScroll === undefined) {
            options.infiniteScroll = true;
        }
        if (options.showButton === undefined) {
            options.showButton = true;
        }
    };

    $.fn.extend({
        // displays a modal filtered picker when the trigger element is clicked
        bindModalFilteredPicker: function (options) {
            cleanOptions(options);
            var trigger = this;
            trigger.click(function() {
                if (options.isDisabled && options.isDisabled()) {
                    return this;
                }

                if (!options.pageSize) {
                    options.pageSize = 20;
                }

                var onClose = function() {
                    dialog.remove();
                    options.onClose && options.onClose();
                };

                var dialog = new AJS.Dialog({ width: 500, height: 607, id: 'modal-filter-picker', keypressListener: function(e) {
                    if(e.keyCode === $.ui.keyCode.ESCAPE) {
                        //Get us out of the current event loop because on FF Ajax may fail since
                        //we're processing an ESC key
                        setTimeout(onClose,0);
                    }
                }});
                dialog.addHeader(options.title, 'filter-dialog-title');
                dialog.addButton(options.closeButtonText, onClose);
                var div = $('<div>').filteredPicker(options);
                dialog.addPanel('filtered-picker-panel', div, options.panelClass);
                div.find('.filter-picker-filter').focus();
                dialog.show();
            });
            return this;
        },
        filteredPicker: function (options) {

            // used to check whether a retrieveItems callback has been cancelled
            var isCancelled = function(candidateCallback) {
                return false;
            };

            if ($(this).length < 1) {
                // nothing to add the filter picker to
                return;
            }

            cleanOptions(options);

            // maintains whether we are currently loading users
            var loading = false;
            // maintains whether the currently displayed list of users is exhaustive for the current filter query
            var maxedOut = false;
            // maintains the number of items clicked since load. We request more after a certain number
            var itemsAddedSinceLoad = 0;
            // some paging APIs return the number offset of the start of the next page. The callback passed to
            // retrieveItems optionally accepts an offset value to be used on the next request.
            var startOfNextPage = null;

            var filteredPickerContainer = this;

            var filterPickerDiv = $(
                '<div class="filter-picker">' +
                '</div>'
            );

            // set up paging offset
            var listOffset = 0;

            var filterLabel = options.filterLabel ? '<label for="filter-picker-filter">' + options.filterLabel + '</label>' : '';

            var filterContainer;
            // add the filter input
            if (options.showFilter) {
                filterContainer = $(
                    "<div class='filter-container disabled'>" + filterLabel +
                        "<input type='text' class='filter-picker-filter' id='filter-picker-filter'>" +
                    "</div>");

                filterPickerDiv.append(filterContainer);
            }

            // set up the filter picker list
            var rowList = $("<ul class='filter-picker-list' />");
            filterPickerDiv.append($("<div class='filter-picker-list-container' />")
                                    .append($("<div class='filter-picker-list-container-inner' />")
                                    .append(rowList)));

            var appendNoItemsMessage = function() {
                rowList.append($('<li class="filter-picker-item no-items">').text(options.noItemsText));
            };

            var appendNoMoreItemsMessage = function() {
                rowList.append($("<li class='filter-picker-item no-items'>").text(options.noMoreItemsText));
            };

            var hideFilter = function() {
                if (filterContainer) {
                    filterContainer.addClass('disabled');
                }
            };

            var showFilter = function() {
                if (filterContainer) {
                    filterContainer.removeClass('disabled');
                }
            };

            var $moreItemsRow;

            var appendMoreItemsButton = function() {
                $moreItemsRow = $('<li class="filter-picker-item more-items">').append(
                    $(stash.buttons.button({
                        extraClasses : 'load-more',
                        buttonText : options.buttonLoadMoreText || stash_i18n('stash.web.filteredpicker.loadmore', 'Load more')
                    })).click(function() {
                        retrieveRows(filter.val(), false);
                    })
                );
                rowList.append($moreItemsRow);
            };

            // function to retrieve item rows in the table
            var retrieveRows = function(filterText, clearOldRows) {
                loading = true;

                var $loadingRow = $('<li class="filter-picker-item loading">').text(options.loadingItemsText || 'Loading');

                var callback = function(items, hasMoreItems, newStartOfNextPage) {
                    // a new request has been sent, drop the results on the floor
                    if (isCancelled(callback)) {
                        return;
                    }

                    // store the start of the next page, if specified
                    if (newStartOfNextPage) {
                        startOfNextPage = newStartOfNextPage;
                    }

                    // reset
                    itemsAddedSinceLoad = 0;

                    var onButtonPress = function(id) {
                        return function() {
                            var $this = $(this);
                            $this.addClass(options.buttonBusyClass)
                                .addClass('filter-picker-busy');
                            var doneCallback = function() {
                                setTimeout(function () {
                                    $this.addClass(options.buttonDoneClass)
                                         .addClass('filter-picker-done')
                                         .text(options.buttonDoneText);
                                    var $parentItem = $this.parent('.filter-picker-item');
                                    $parentItem.delay(1000)
                                        .slideUp("fast", "swing", function() {
                                            var doShowFilter = true;

                                            $parentItem.remove();
                                            if (filterPickerDiv.find("li.filter-picker-item").size() === 0) {
                                                if (options.showFilter && filterPickerDiv.find(".filter-picker-filter").val() === "") {
                                                  // we're not filtering and there are no items
                                                  appendNoMoreItemsMessage();
                                                  doShowFilter = false;
                                                } else {
                                                  // we're filtering and there are no items
                                                  appendNoItemsMessage();
                                                }
                                            } else if (++itemsAddedSinceLoad >= 10 && !maxedOut) {
                                                // if 10 or more items have been added, retrieve some more items
                                                retrieveRows(filter.val(), false);
                                            }

                                            if (doShowFilter) {
                                                showFilter();
                                            } else {
                                                hideFilter();
                                            }
                                        });
                                }, 500);
                            };
                            var errorCallback = function(errors) {
                                $this.addClass('disabled').text(options.buttonText || '');
                                var parent = $this.closest('.filter-picker-item');
                                $.each(errors, function(i, error) {
                                    parent.append($('<span class="error-message" />').text(error.message));
                                });
                            };
                            $this.unbind("click").text(options.buttonDoingText);
                            options.buttonPressed(id, doneCallback, errorCallback);

                            // decrement offset as the clicked entity will no longer be part of the returned list
                            listOffset--;
                        };
                    };

                    // clear the old more-items button if it exists
                    $moreItemsRow && $moreItemsRow.remove();
                    $loadingRow && $loadingRow.remove();

                    // add new results
                    for (var i = 0, l = items.length; i < l; i++) {
                        var item = items[i];
                        var li = $("<li class='filter-picker-item' tabindex='0'>");
                        if (i === 0) {
                            li.addClass("top");
                        }
                        li.attr("data-id", item.id); // for ease of web-driver testing
                        li.append(item.content.addClass("filter-picker-item-content"));

                        if (options.showButton) {
                            var button = $(stash.buttons.button({
                                extraClasses : 'filter-picker-item-button',
                                buttonText : options.buttonText || '',
                                isDisabled: item.disabled,
                                title: item.disabled ? item.disabledReason : null
                            }));

                            if (!item.disabled) {
                                button.click(onButtonPress(item.id));
                            }
                            li.append(button);
                        }

                        rowList.append(li);
                    }
                    listOffset += items.length;

                    var doShowFilter = true;
                    if (items.length === 0 || !hasMoreItems) {
                        // no more items and the filter input is empty
                        if (items.length === 0 && !hasMoreItems && options.showFilter && filterPickerDiv.find(".filter-picker-filter").val() === "") {
                            doShowFilter = false;
                            appendNoMoreItemsMessage();
                        } else if (filterPickerDiv.find("li.filter-picker-item").size() === 0) {
                            appendNoItemsMessage();
                        }
                        maxedOut = true;
                    } else if (!options.infiniteScroll) {
                        appendMoreItemsButton();
                    }

                    if (doShowFilter) {
                        showFilter();
                    } else {
                        hideFilter();
                    }

                    loading = false;
                    itemsAddedSinceLoad = 0; // reset items added counter
                };
                if (clearOldRows) {
                    // clear out old items
                    rowList.find('.filter-picker-item:not(.filter-container)').remove();
                    listOffset = 0;
                    startOfNextPage = null;
                    maxedOut = false;
                }

                $moreItemsRow && $moreItemsRow.remove();
                rowList.append($loadingRow);
                // redefine isCancelled to return true for all other callbacks, we should drop responses to earlier queries on the floor
                isCancelled = function(candidateCallback) {
                    return candidateCallback !== callback;
                };
                options.retrieveItems(filterText, startOfNextPage ? startOfNextPage : listOffset, options.pageSize, callback);
            };

            var filter;
            if (options.showFilter) {
                // set up the filter field
                filter = filterPickerDiv.find('.filter-picker-filter');
                // use the jquery debounce plugin to prevent requests being fired too quickly/often
                filter.keyup(_.debounce(function() {
                    retrieveRows(filter.val(), true);
                }, 350));
            } else {
                // todo fix this hack
                filter = {
                    val: function() {
                        return "";
                    }
                };
            }

            //TODO - use PagedScrollable.
            if (options.infiniteScroll) {
                filterPickerDiv.find('.filter-picker-list-container').scroll(function() {
                    var $this = $(this);
                    var $inner = filterPickerDiv.find('.filter-picker-list-container-inner');
                    if (!loading && !maxedOut &&
                            // the ten pixel buffer is due to a height calculation issue in IE8
                            $inner.outerHeight() > 10 && // prevent the hack from causing aggressive scrolling
                            Math.abs($inner.offset().top) + $this.height() + $this.offset().top >= $inner.outerHeight() - 10) {
                        // trigger infinite scroll
                        retrieveRows(filter.val(), false);
                    }
                });
            }

            filteredPickerContainer.append(filterPickerDiv);

            // initialise with a search for the empty string (should return all values)
            retrieveRows("", false);

            return $.extend(this, {
                refreshPicker: function() {
                    retrieveRows(filter.val(), true);
                }
            });
        }
    });
})(jQuery, _);
